from enum import Enum
from typing import Optional

from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.inline_link import InlineLink
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.rst.rst_templates import RSTTemplates
from strictdoc.helpers.rst import escape_str_after_inline_markup


class TAG(Enum):
    SECTION = 1
    REQUIREMENT = 2
    COMPOSITE_REQUIREMENT = 3


class RSTWriter:
    def __init__(self, index: TraceabilityIndex):
        self.index = index

    def write(self, document: SDocDocument, single_document: bool) -> str:
        document_iterator = DocumentCachingIterator(document)
        output = ""

        if not single_document:
            document_uid: Optional[str] = None
            if document.config.uid is not None:
                document_uid = document.config.uid
            output += self._print_rst_header(
                document.title, 0, reference_uid=document_uid
            )

        for content_node, _ in document_iterator.all_content():
            if isinstance(content_node, SDocSection):
                assert content_node.ng_level is not None
                output += self._print_rst_header(
                    content_node.title,
                    content_node.ng_level,
                    content_node.reserved_uid,
                )

            elif isinstance(content_node, SDocNode):
                output += self._print_requirement_fields(content_node)

        if output.endswith("\n\n"):
            output = output[:-1]
        return output.lstrip()

    @staticmethod
    def _print_rst_header_2(string: str, level: int) -> str:
        assert isinstance(string, str), string
        assert isinstance(level, int), level
        chars = {
            0: "$",
            1: "=",
            2: "-",
            3: "~",
            4: "^",
            5: '"',
            6: "#",
            7: "'",
        }
        header_char = chars[level]
        output = ""
        output += string
        output += "\n"
        output += header_char.rjust(len(string), header_char)
        return output

    @staticmethod
    def _print_rst_header(
        string: str, level: int, reference_uid: Optional[str]
    ) -> str:
        assert isinstance(string, str), string
        assert isinstance(level, int), level
        chars = {
            0: "$",
            1: "=",
            2: "-",
            3: "~",
            4: "^",
            5: '"',
            6: "#",
            7: "'",
        }
        header_char = chars[level]
        output = ""

        # An RST reference looks like this:
        # .. _SDOC-HIGH-VALIDATION:
        if reference_uid is not None:
            assert len(reference_uid) > 0, reference_uid
            output += f".. _{reference_uid}:\n\n"

        output += string
        output += "\n"
        output += header_char.rjust(len(string), header_char)
        output += "\n\n"
        return output

    def _print_requirement_fields(self, section_content: SDocNode) -> str:
        requirement_template = RSTTemplates.jinja_environment.get_template(
            "requirement.jinja.rst"
        )
        output = requirement_template.render(
            requirement=section_content,
            index=self.index,
            _print_rst_header=self._print_rst_header_2,
            _print_node_field=self._print_node_field,
        )
        return output

    def _print_node_field(self, object_with_parts: SDocNodeField) -> str:
        output = ""
        prev_part = None
        for part in object_with_parts.parts:
            if isinstance(part, str):
                if isinstance(prev_part, InlineLink):
                    output += escape_str_after_inline_markup(part)
                else:
                    output += part
            elif isinstance(part, InlineLink):
                node_or_none = self.index.get_linkable_node_by_uid(part.link)
                # Labels that aren't placed before a section title can still be
                # referenced, but you must give the link an explicit title,
                # using this syntax: :ref:`Link title <label-name>`.
                # https://www.sphinx-doc.org/en/master/usage/restructuredtext/roles.html
                node_display_title = node_or_none.get_display_title(
                    include_toc_number=False
                )
                output += f":ref:`{node_display_title} <{part.link}>`"
            elif isinstance(part, Anchor):
                output += f".. _{part.value}:\n"
            else:
                raise NotImplementedError
            prev_part = part

        return output



from typing import Any, Dict, List, Optional, Tuple

from docutils import nodes
from docutils.parsers.rst import Directive, directives
from docutils.parsers.rst.states import Inliner


def eq_role(
    _role: str,
    _rawtext: str,
    text: str,
    _lineno: int,
    _inliner: Inliner,
    _options: Optional[Dict[str, Any]] = None,
    _content: Optional[List[str]] = None,
) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
    """Handle rst :eq: role, producing a link to an equation."""
    eqref = f"\\( \\eqref{{{text}}} \\)"
    node = nodes.raw(
        "",
        f'<span class="math notranslate nohighlight">{eqref}</span>',
        format="html",
    )
    return [node], []


def eq_role_for_server(
    _role: str,
    _rawtext: str,
    text: str,
    _lineno: int,
    _inliner: Inliner,
    _options: Optional[Dict[str, Any]] = None,
    _content: Optional[List[str]] = None,
) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
    """Handle rst :eq: role, producing a link to an equation. (server edition)."""
    eqref = f"\\( \\eqref{{{text}}} \\)"

    node = nodes.raw(
        "",
        f'<span class="math notranslate nohighlight" original-math="{eqref}">{eqref}</span>',
        format="html",
    )
    return [node], []


def math_role(
    _role: str,
    rawtext: str,
    text: str,
    _lineno: int,
    _inliner: Inliner,
    _options: Optional[Dict[str, Any]] = None,
    _content: Optional[List[str]] = None,
) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
    """Handle rst :math: role, rendering a formula inline with text."""
    if rawtext.startswith(":math:`") and rawtext.endswith("`"):
        extracted = rawtext[len(":math:`") : -1]
    else:
        extracted = text

    extracted = f"\\( {extracted} \\)"

    node = nodes.raw(
        "",
        f'<span class="math notranslate nohighlight">{extracted}</span>',
        format="html",
    )
    return [node], []


def math_role_for_server(
    _role: str,
    rawtext: str,
    text: str,
    _lineno: int,
    _inliner: Inliner,
    _options: Optional[Dict[str, Any]] = None,
    _content: Optional[List[str]] = None,
) -> Tuple[List[nodes.Node], List[nodes.system_message]]:
    """Handle rst :math: role, rendering a formula inline with text. (server edition)."""
    if rawtext.startswith(":math:`") and rawtext.endswith("`"):
        extracted = rawtext[len(":math:`") : -1]
    else:
        extracted = text

    extracted = f"\\( {extracted} \\)"

    node = nodes.raw(
        "",
        f'<span class="math notranslate nohighlight" original-math="{extracted}">{extracted}</span>',
        format="html",
    )
    return [node], []


class MathDirective(Directive):  # type: ignore[misc]
    """Docutils directive emulating Sphinx-style math syntax."""

    is_running_in_server: bool = False
    has_content = True
    optional_arguments = 0
    final_argument_whitespace = False
    option_spec = {
        "nowrap": directives.flag,
        "no-wrap": directives.flag,
        "label": directives.unchanged,
        "number": directives.flag,
    }

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Constuct a MathDirective."""
        self.is_running_in_server = False
        super().__init__(*args, **kwargs)

    def run(self) -> List[nodes.Node]:
        """Handle ..math :: directive, rendering a formula centered, optionally with eq number."""

        nowrap = "no-wrap" in self.options or "nowrap" in self.options
        label = self.options.get("label")

        math_content = "\n".join(self.content).strip()
        html_parts = []

        if nowrap:
            mathjax_content = math_content
        else:
            parts = [p.strip() for p in math_content.split("\n\n") if p.strip()]
            if len(parts) > 1:
                wrapped = []
                for part in parts:
                    if r"\\" in part:
                        wrapped.append(r"\begin{split}" + part + r"\end{split}")
                    else:
                        wrapped.append(part)

                if label:
                    mathjax_content = (
                        f"\\begin{{equation}}\\label{{{label}}}"
                        r"\begin{aligned}"
                        + r" \\ ".join(wrapped)
                        + r"\end{aligned}\end{equation}"
                    )
                else:
                    mathjax_content = (
                        r"\begin{align*}\begin{aligned}"
                        + r" \\ ".join(wrapped)
                        + r"\end{aligned}\end{align*}"
                    )
            else:
                if label:
                    mathjax_content = f"\\begin{{equation}}\\label{{{label}}}\\begin{{aligned}} {math_content} \\end{{aligned}}\\end{{equation}}"
                else:
                    mathjax_content = (
                        f" \\begin{{align*}}  {math_content} \\end{{align*}}"
                    )

        mathjax_content = f"\\[ {mathjax_content} \\]"

        if self.is_running_in_server:
            html_parts.append(
                f'<div class="math notranslate nohighlight" original-math="{mathjax_content}">'
            )
        else:
            html_parts.append('<div class="math notranslate nohighlight">')
        html_parts.append(mathjax_content)
        html_parts.append("</div>")

        html = "".join(html_parts)
        return [nodes.raw("", html, format="html")]


class MathDirectiveForServer(MathDirective):
    """Docutils directive emulating Sphinx-style math syntax. (server edition)"""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Constuct a MathDirective, and set flag for running in server-mode."""
        super().__init__(*args, **kwargs)
        self.is_running_in_server = True



"""
@relation(SDOC-SRS-33, scope=file)
"""

import os
from enum import Enum
from typing import Dict, List, Sequence

from strictdoc.core.file_tree import File, FileTree
from strictdoc.helpers.auto_described import auto_described


class SourceFileType(Enum):
    PYTHON = [".py"]
    C = [".h", ".c"]
    CPP = [".cpp", ".cc"]
    TEX = [".tex"]
    # Is there an idiomatic file extension for Jinja templates?
    # https://stackoverflow.com/questions/29590931/is-there-an-idiomatic-file-extension-for-jinja-templates
    TOML = [".toml"]
    JINJA = [".jinja", ".jinja2", ".j2", ".html.jinja"]
    JAVASCRIPT = [".js"]
    YAML = [".yaml", ".yml"]
    RST = [".rst"]
    SDOC = [".sdoc"]

    UNKNOWN: List[str] = []  # type: ignore[misc]

    @classmethod
    def create_from_path(cls, path_to_file: str) -> "SourceFileType":
        if path_to_file.endswith(".py"):
            return cls.PYTHON
        if path_to_file.endswith(".h") or path_to_file.endswith(".c"):
            return cls.C
        for enum_value in SourceFileType.CPP.value:
            if path_to_file.endswith(enum_value):
                return cls.CPP
        if path_to_file.endswith(".tex"):
            return cls.TEX
        if path_to_file.endswith(".toml"):
            return cls.TOML
        for enum_value in SourceFileType.JINJA.value:
            if path_to_file.endswith(enum_value):
                return cls.JINJA
        if path_to_file.endswith(".js"):
            return cls.JAVASCRIPT
        for enum_value in SourceFileType.YAML.value:
            if path_to_file.endswith(enum_value):
                return cls.YAML
        for enum_value in SourceFileType.RST.value:
            if path_to_file.endswith(enum_value):
                return cls.RST
        if path_to_file.endswith(".sdoc"):
            return cls.SDOC

        return cls.UNKNOWN


@auto_described
class SourceFile:
    def __init__(
        self,
        level: int,
        full_path: str,
        in_doctree_source_file_rel_path: str,
        output_dir_full_path: str,
        output_file_full_path: str,
    ) -> None:
        assert isinstance(level, int)
        assert os.path.exists(full_path)

        self.level = level
        self.full_path = full_path
        self.file_name = os.path.basename(full_path)
        self.in_doctree_source_file_rel_path = in_doctree_source_file_rel_path
        self.in_doctree_source_file_rel_path_posix: str = (
            in_doctree_source_file_rel_path.replace("\\", "/")
        )
        self.output_dir_full_path = output_dir_full_path
        self.output_file_full_path = output_file_full_path
        self.path_depth_prefix = ("../" * (level + 1))[:-1]

        self.file_type: SourceFileType = SourceFileType.create_from_path(
            in_doctree_source_file_rel_path
        )

        self.is_referenced = False

    def is_python_file(self) -> bool:
        return self.file_type == SourceFileType.PYTHON

    def is_c_file(self) -> bool:
        return self.file_type == SourceFileType.C

    def is_cpp_file(self) -> bool:
        return self.file_type == SourceFileType.CPP

    def is_tex_file(self) -> bool:
        return self.file_type == SourceFileType.TEX

    def is_toml_file(self) -> bool:
        return self.file_type == SourceFileType.TOML

    def is_jinja_file(self) -> bool:
        return self.file_type == SourceFileType.JINJA

    def is_javascript_file(self) -> bool:
        return self.file_type == SourceFileType.JAVASCRIPT

    def is_yaml_file(self) -> bool:
        return self.file_type == SourceFileType.YAML

    def is_rst_file(self) -> bool:
        return self.file_type == SourceFileType.RST


class SourceTree:
    def __init__(
        self,
        file_tree: FileTree,
        source_files: Sequence[SourceFile],
        map_file_to_source: Dict[File, SourceFile],
    ):
        self.file_tree: FileTree = file_tree
        self.source_files: Sequence[SourceFile] = source_files
        self.map_file_to_source: Dict[File, SourceFile] = map_file_to_source

    def get_source_for_file(self, file: File) -> SourceFile:
        assert isinstance(file, File)
        return self.map_file_to_source[file]



import os
import platform
import shutil
import tempfile
from pathlib import Path
from typing import Optional


def sync_dir(src_dir: str, dst_dir: str, message: Optional[str]) -> None:
    assert os.path.isabs(src_dir), f"Expected {src_dir} to be an absolute path"
    assert os.path.isdir(src_dir), f"Expected {src_dir} to be a directory"

    Path(dst_dir).mkdir(parents=True, exist_ok=True)

    # First, collect all source files to be copied.
    source_files = []
    for root, _, filenames in os.walk(src_dir, topdown=True):
        for filename in filenames:
            full_path_to_file = os.path.join(root, filename)
            relative_path_to_file_dir = os.path.relpath(root, start=src_dir)
            if relative_path_to_file_dir == ".":
                relative_path_to_file_dir = ""
            source_files.append(
                (
                    filename,
                    full_path_to_file,
                    relative_path_to_file_dir,
                )
            )

    if len(source_files) == 0:
        return

    # Print progress message.
    if message is not None:
        print(f"{message}: {len(source_files)} files", end="")  # noqa: T201

    # Iterate through the files to be copied and copy them one by one.
    skipped = 0
    for (
        filename,
        full_path_to_file,
        relative_path_to_file_dir,
    ) in source_files:
        full_path_to_destination_dir = os.path.join(
            dst_dir, relative_path_to_file_dir
        )
        Path(full_path_to_destination_dir).mkdir(parents=True, exist_ok=True)
        full_path_to_destination_file = os.path.join(
            full_path_to_destination_dir, filename
        )

        # StrictDoc: We don't want to copy files if they are not newer.
        if not os.path.isfile(
            full_path_to_destination_file
        ) or os.path.getmtime(full_path_to_file) > os.path.getmtime(
            full_path_to_destination_file
        ):
            shutil.copyfile(full_path_to_file, full_path_to_destination_file)
        else:
            skipped += 1

    if message is not None:
        if skipped > 0:
            print(  # noqa: T201
                f" ({skipped} files skipped as non-modified).", flush=True
            )
        else:
            print(".", flush=True)  # noqa: T201


def get_portable_temp_dir() -> Path:
    return Path(
        "/tmp" if platform.system() == "Darwin" else tempfile.gettempdir()
    )



from dataclasses import dataclass
from typing import Any, Dict, List

from markupsafe import Markup

from strictdoc.export.html.html_templates import JinjaEnvironment


@dataclass
class RowWithReservedFieldFormObject:
    field: Any
    errors: Dict[str, List[str]]
    jinja_environment: JinjaEnvironment

    def __post_init__(self) -> None:
        assert isinstance(self.jinja_environment, JinjaEnvironment), (
            self.jinja_environment
        )

    def render(self) -> Markup:
        rendered_template = self.jinja_environment.render_template_as_markup(
            "components/grammar_form_element/row_with_reserved_field/index.jinja",
            form_object=self,
        )
        return rendered_template

    def get_errors(self, field_name: str) -> List[str]:
        return self.errors.get(field_name, [])


@dataclass
class RowWithCustomFieldFormObject:
    field: Any
    errors: Dict[str, List[str]]
    jinja_environment: JinjaEnvironment

    def __post_init__(self) -> None:
        assert self.field is not None
        assert isinstance(self.jinja_environment, JinjaEnvironment), (
            self.jinja_environment
        )

    def render(self) -> Markup:
        rendered_template = self.jinja_environment.render_template_as_markup(
            "components/grammar_form_element/row_with_custom_field/index.jinja",
            form_object=self,
        )
        return rendered_template

    def get_errors(self, field_name: str) -> List[str]:
        return self.errors.get(field_name, [])


@dataclass
class RowWithRelationFormObject:
    relation: Any
    errors: Dict[str, List[str]]
    jinja_environment: JinjaEnvironment

    def __post_init__(self) -> None:
        assert self.relation is not None
        assert isinstance(self.jinja_environment, JinjaEnvironment), (
            self.jinja_environment
        )

    def render(self) -> Markup:
        rendered_template = self.jinja_environment.render_template_as_markup(
            "components/grammar_form_element/row_with_relation/index.jinja",
            form_object=self,
        )
        return rendered_template

    def get_errors(self, field_name: str) -> List[str]:
        return self.errors.get(field_name, [])



"""
@relation(SDOC-SRS-33, scope=file)
"""

from typing import List, Optional, Tuple

from strictdoc.backend.sdoc.error_handling import StrictDocSemanticError
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.backend.sdoc_source_code.parse_context import ParseContext


def source_file_traceability_info_processor(
    source_file_traceability_info: SourceFileTraceabilityInfo,
    parse_context: ParseContext,
) -> None:
    if len(parse_context.marker_stack) > 0:
        raise create_unmatch_range_error(
            parse_context.marker_stack, parse_context.filename
        )
    source_file_traceability_info.markers = parse_context.markers
    source_file_traceability_info.file_stats = parse_context.file_stats
    source_file_traceability_info.ng_map_reqs_to_markers = (
        parse_context.map_reqs_to_markers
    )


def create_begin_end_range_reqs_mismatch_error(
    filename: str,
    line: int,
    col: int,
    lhs_marker_reqs: List[str],
    rhs_marker_reqs: List[str],
) -> StrictDocSemanticError:
    lhs_marker_reqs_str = ", ".join(lhs_marker_reqs)
    rhs_marker_reqs_str = ", ".join(rhs_marker_reqs)

    return StrictDocSemanticError(
        "STRICTDOC RANGE: BEGIN and END requirements mismatch",
        (
            "STRICT RANGE marker should START and END "
            "with the same requirement(s): "
            f"'{lhs_marker_reqs_str}' != '{rhs_marker_reqs_str}'."
        ),
        # @sdoc[nosdoc]  # noqa: ERA001
        """
# [REQ-001]
Content...
# [/REQ-001]
        """.lstrip(),
        # @sdoc[/nosdoc]  # noqa: ERA001
        line=line,
        col=col,
        filename=filename,
    )


def create_end_without_begin_error(
    filename: str, line: int, col: int
) -> StrictDocSemanticError:
    return StrictDocSemanticError(
        "STRICTDOC RANGE: END marker without preceding BEGIN marker",
        (
            "STRICT RANGE shall be opened with "
            "START marker and ended with END marker."
        ),
        # @sdoc[nosdoc]  # noqa: ERA001
        """
# [REQ-001]
Content...
# [/REQ-001]
        """.lstrip(),
        # @sdoc[/nosdoc]  # noqa: ERA001
        line=line,
        col=col,
        filename=filename,
    )


def create_unmatch_range_error(
    unmatched_ranges: List[RangeMarker], filename: Optional[str]
) -> StrictDocSemanticError:
    assert isinstance(unmatched_ranges, list)
    assert len(unmatched_ranges) > 0
    range_locations: List[Tuple[int, int]] = []
    for unmatched_range_ in unmatched_ranges:
        assert unmatched_range_.ng_source_line_begin is not None
        assert unmatched_range_.ng_source_column_begin is not None
        range_locations.append(
            (
                unmatched_range_.ng_source_line_begin,
                unmatched_range_.ng_source_column_begin,
            )
        )
    first_location = range_locations[0]
    hint: Optional[str] = None
    if len(unmatched_ranges) > 1:
        range_lines = range_locations[1:]
        hint = f"The @sdoc keywords are also unmatched on lines: {range_lines}."

    return StrictDocSemanticError(
        "Unmatched @sdoc keyword found in source file.",
        hint=hint,
        # @sdoc[nosdoc]
        example=(
            "Each @sdoc keyword must be matched with a closing keyword. "
            "Example:\n"
            "@sdoc[REQ-001]\n"
            "...\n"
            "@sdoc[/REQ-001]"
        ),
        # @sdoc[/nosdoc]
        line=first_location[0],
        col=first_location[1],
        filename=filename,
    )


def function_range_marker_processor(
    marker: FunctionRangeMarker, parse_context: ParseContext
) -> None:
    if marker.ng_is_nodoc:
        return

    if (
        len(parse_context.marker_stack) > 0
        and parse_context.marker_stack[-1].ng_is_nodoc
    ):
        # This marker is within a "nosdoc" block, so we ignore it.
        return

    parse_context.markers.append(marker)

    assert marker.ng_source_line_begin is not None
    for req in marker.reqs:
        markers = parse_context.map_reqs_to_markers.setdefault(req, [])
        markers.append(marker)


def range_marker_processor(
    marker: RangeMarker, parse_context: ParseContext
) -> None:
    current_top_marker: RangeMarker

    if marker.ng_is_nodoc:
        if marker.is_begin():
            parse_context.marker_stack.append(marker)
        elif marker.is_end():
            try:
                current_top_marker = parse_context.marker_stack.pop()
                if (
                    not current_top_marker.ng_is_nodoc
                    or current_top_marker.is_end()
                ):
                    raise create_begin_end_range_reqs_mismatch_error(
                        "FIXME", -1, -1, current_top_marker.reqs, marker.reqs
                    )
            except IndexError:
                raise create_end_without_begin_error("FIXME", -1, -1) from None
        return

    if (
        len(parse_context.marker_stack) > 0
        and parse_context.marker_stack[-1].ng_is_nodoc
    ):
        # This marker is within a "nosdoc" block, so we ignore it.
        return

    parse_context.markers.append(marker)

    assert marker.ng_source_line_begin is not None

    if marker.is_begin():
        marker.ng_range_line_begin = marker.ng_source_line_begin
        parse_context.marker_stack.append(marker)
        assert marker.ng_source_line_begin is not None
        for req in marker.reqs:
            markers = parse_context.map_reqs_to_markers.setdefault(req, [])
            markers.append(marker)

    elif marker.is_end():
        try:
            current_top_marker = parse_context.marker_stack.pop()
            if marker.reqs != current_top_marker.reqs:
                assert marker.ng_source_line_begin is not None
                raise create_begin_end_range_reqs_mismatch_error(
                    "FIXME",
                    marker.ng_source_line_begin,
                    -1,
                    current_top_marker.reqs,
                    marker.reqs,
                )

            current_top_marker.ng_range_line_end = marker.ng_source_line_begin

            marker.ng_range_line_end = marker.ng_range_line_begin
            marker.ng_range_line_begin = current_top_marker.ng_range_line_begin

        except IndexError:
            raise create_end_without_begin_error("FIXME", 1, 1) from None
    else:
        raise NotImplementedError


def line_marker_processor(
    line_marker: LineMarker, parse_context: ParseContext
) -> None:
    if (
        len(parse_context.marker_stack) > 0
        and parse_context.marker_stack[-1].ng_is_nodoc
    ):
        # This marker is within a "nosdoc" block, so we ignore it.
        return

    parse_context.markers.append(line_marker)

    assert line_marker.ng_source_line_begin is not None

    for req in line_marker.reqs:
        markers = parse_context.map_reqs_to_markers.setdefault(req, [])
        markers.append(line_marker)



GRAMMAR_GRAMMAR = r"""
DocumentGrammar[noskipws]:
  '[GRAMMAR]' '\n'

  (
    (
      'ELEMENTS:' '\n'
      elements += GrammarElement
    )? |
    (
      'IMPORT_FROM_FILE: ' import_from_file = /.+$/ '\n'
    )?
  )
;

GrammarElement[noskipws]:
  '- TAG: ' tag = RequirementType '\n'
  (
    '  PROPERTIES:' '\n'
    ('    IS_COMPOSITE: ' property_is_composite=/(True|False)/ '\n' )?
    ('    PREFIX: ' property_prefix=/.*/ '\n' )?
    ('    VIEW_STYLE: ' property_view_style=/(Plain|Simple|Inline|Narrative|Table|Zebra)/ '\n')?
  )?
  '  FIELDS:' '\n'
  fields += GrammarElementField
  (
    '  RELATIONS:' '\n'
    relations += GrammarElementRelation
  )?
;

GrammarElementRelation[noskipws]:
  (GrammarElementRelationParent | GrammarElementRelationChild | GrammarElementRelationFile)
;

GrammarElementRelationParent[noskipws]:
  '  - TYPE: ' relation_type='Parent' '\n'
  ('    ROLE: ' relation_role=/.+/ '\n')?
;

GrammarElementRelationChild[noskipws]:
  '  - TYPE: ' relation_type='Child' '\n'
  ('    ROLE: ' relation_role=/.+/ '\n')?
;

GrammarElementRelationFile[noskipws]:
  '  - TYPE: ' relation_type='File' '\n'
  ('    ROLE: ' relation_role=/.+/ '\n')?
;

GrammarElementField[noskipws]:
  GrammarElementFieldString |
  GrammarElementFieldSingleChoice |
  GrammarElementFieldMultipleChoice |
  GrammarElementFieldTag
;

GrammarElementFieldString[noskipws]:
  '  - TITLE: ' title=FieldName '\n'
  ('    HUMAN_TITLE: ' human_title=SingleLineString '\n')?
  '    TYPE: String' '\n'
  '    REQUIRED: ' (required = BooleanChoice) '\n'
;

GrammarElementFieldSingleChoice[noskipws]:
  '  - TITLE: ' title=FieldName '\n'
  ('    HUMAN_TITLE: ' human_title=SingleLineString '\n')?
  '    TYPE: SingleChoice'
    '(' ((options = ChoiceOption) (options *= ChoiceOptionXs)) ')' '\n'
  '    REQUIRED: ' (required = BooleanChoice) '\n'
;

GrammarElementFieldMultipleChoice[noskipws]:
  '  - TITLE: ' title=FieldName '\n'
  ('    HUMAN_TITLE: ' human_title=SingleLineString '\n')?
  '    TYPE: MultipleChoice'
    '(' ((options = ChoiceOption) (options *= ChoiceOptionXs)) ')' '\n'
  '    REQUIRED: ' (required = BooleanChoice) '\n'
;

GrammarElementFieldTag[noskipws]:
  '  - TITLE: ' title=FieldName '\n'
  ('    HUMAN_TITLE: ' human_title=SingleLineString '\n')?
  '    TYPE: Tag' '\n'
  '    REQUIRED: ' (required = BooleanChoice) '\n'
;

ReferenceType[noskipws]:
  ('ParentReqReference' | 'ChildReqReference' | 'FileReference')
;

ReferenceTypeXs[noskipws]:
  /, /- ReferenceType
;
"""

GRAMMAR_WRAPPER = """
DocumentGrammarWrapper[noskipws]:
    grammar = DocumentGrammar
;
"""



import os
import sys
import time
from typing import Awaitable, Callable

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from starlette.requests import Request
from starlette.responses import Response

from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.coverage import register_code_coverage_hook
from strictdoc.helpers.pickle import pickle_load
from strictdoc.server.config import SDocServerEnvVariable
from strictdoc.server.routers.main_router import create_main_router
from strictdoc.server.routers.other_router import create_other_router

# Define O_TEMPORARY for Windows only
if sys.platform == "win32":
    O_TEMPORARY = os.O_TEMPORARY  # pragma: no cover
else:
    O_TEMPORARY = 0


def create_app(*, project_config: ProjectConfig) -> FastAPI:
    app = FastAPI()

    origins = [
        "http://localhost",
        "http://localhost:8081",
        "http://localhost:3000",
    ]

    # Uncomment this to enable performance measurements.
    @app.middleware("http")
    async def add_process_time_header(  # pylint: disable=unused-variable
        request: Request, call_next: Callable[[Request], Awaitable[Response]]
    ) -> Response:
        start_time = time.time()
        response: Response = await call_next(request)
        time_passed = round(time.time() - start_time, 3)

        request_path = request.url.path
        if len(request.url.query) > 0:
            request_path += f"?{request.url.query}"

        print(  # noqa: T201
            f"PERF:     {request.method} {request_path} {time_passed}s"
        )
        return response

    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    app.include_router(create_other_router(project_config=project_config))
    app.include_router(create_main_router(project_config=project_config))

    return app


def strictdoc_production_app() -> FastAPI:
    register_code_coverage_hook()

    # This is a work-around to allow opening a file created with
    # NamedTemporaryFile on Windows.
    # See https://stackoverflow.com/a/15235559
    def temp_opener(name: str, flag: int, mode: int = 0o777) -> int:
        flag |= O_TEMPORARY
        return os.open(name, flag, mode)

    path_to_tmp_config = os.environ[SDocServerEnvVariable.PATH_TO_CONFIG]
    with open(path_to_tmp_config, "rb", opener=temp_opener) as tmp_config_file:
        tmp_config_bytes = tmp_config_file.read()

    project_config = pickle_load(tmp_config_bytes)
    assert isinstance(project_config, ProjectConfig), project_config

    return create_app(
        project_config=project_config,
    )



import re
from typing import Dict, List, Match, Optional, Tuple, Union

from typing_extensions import TypeAlias

ParsedFormDataLeaf: TypeAlias = str

ParsedFormDataContainer: TypeAlias = Union[
    Dict[str, Union[ParsedFormDataLeaf, "ParsedFormDataContainer"]],
    List[Union[ParsedFormDataLeaf, "ParsedFormDataContainer"]],
]

ParsedFormData: TypeAlias = Dict[
    str, Union[ParsedFormDataLeaf, ParsedFormDataContainer]
]

ParsedFormDataNode: TypeAlias = Union[
    str,
    Dict[str, "ParsedFormDataNode"],
    List["ParsedFormDataNode"],
]


def as_dict(node: ParsedFormDataNode) -> Dict[str, ParsedFormDataNode]:
    if not isinstance(node, dict):
        raise KeyError("Expected dict node")
    return node


def as_list(node: ParsedFormDataNode) -> List[ParsedFormDataNode]:
    if not isinstance(node, list):
        raise KeyError("Expected list node")
    return node


def _set_value_by_key_path(
    obj: Dict[str, ParsedFormDataNode],
    parts: List[Union[str, int]],
    value: Union[ParsedFormDataLeaf, ParsedFormDataContainer],
) -> None:
    cursor: ParsedFormDataNode = obj
    for part_idx, part in enumerate(parts):
        try:
            next_component = parts[part_idx + 1]

            if isinstance(part, str):
                cursor_dict = as_dict(cursor)

                if isinstance(next_component, str):
                    if next_component != "":
                        if part not in cursor:
                            if not isinstance(cursor, dict):
                                raise KeyError
                            cursor_dict[part] = {}
                    else:
                        if part not in cursor:
                            cursor_dict[part] = []
                else:
                    if part not in cursor:
                        cursor_dict[part] = []

                cursor = cursor_dict[part]
            else:
                assert isinstance(part, int)

                if isinstance(next_component, str):
                    cursor_list = as_list(cursor)

                    try:
                        cursor = cursor_list[part]
                    except IndexError:
                        if part != len(cursor):
                            raise IndentationError(
                                "The ordering [0], [1], ... "
                                "is broken in this form data: "
                                f"{obj} {parts} {value}."
                            ) from None
                        cursor_list.append({})
                        cursor = cursor_list[part]
                else:
                    raise NotImplementedError("No [][] supported.")
        except IndexError:
            if isinstance(part, str):
                assert len(part) > 0, part
                assert isinstance(cursor, dict)
                cursor[part] = value
            else:
                raise NotImplementedError from None  # pragma: no cover


FIELD_NAME = "[A-Za-z0-9_]*"


def parse_form_data(
    form_data: List[Tuple[str, str]],
) -> ParsedFormData:
    result_dict: ParsedFormData = {}

    for key, value in form_data:
        first_match: Optional[Match[str]] = re.match(rf"({FIELD_NAME})", key)

        assert first_match is not None, "first_match must not be None."

        match_groups = first_match.groups()
        assert len(match_groups) == 1

        match = re.findall(rf"\[({FIELD_NAME}|\d+)]", key)
        strings: List[Union[str, int]] = list(map(str, match))
        for part_idx, part in enumerate(strings):
            try:
                strings[part_idx] = int(part)
            except ValueError:
                continue

        components = [match_groups[0]] + strings

        _set_value_by_key_path(result_dict, components, value)

    return result_dict



import difflib
from difflib import SequenceMatcher
from typing import Callable

from markupsafe import Markup, escape


def similar(a: str, b: str) -> float:
    return SequenceMatcher(None, a, b).ratio()


red: Callable[[str], str] = (
    lambda text: f'<span class="lambda_red">{escape(text)}</span>'
)
green: Callable[[str], str] = (
    lambda text: f'<span class="lambda_green">{escape(text)}</span>'
)
white: Callable[[str], str] = lambda text: f"<span>{escape(text)}</span>"


def get_colored_html_diff_string(old: str, new: str, flag: str) -> Markup:
    assert old is not None
    assert new is not None
    assert flag in ("left", "right")

    result = ""
    codes = difflib.SequenceMatcher(a=old, b=new).get_opcodes()
    for code in codes:
        if code[0] == "equal":
            result += white(old[code[1] : code[2]])
        elif code[0] == "delete":
            if flag == "left":
                result += red(old[code[1] : code[2]])
        elif code[0] == "insert":
            if flag == "right":
                result += green(new[code[3] : code[4]])
        elif code[0] == "replace":
            if flag == "left":
                result += red(old[code[1] : code[2]])
            else:
                result += green(new[code[3] : code[4]])
        else:
            raise NotImplementedError(code)  # pragma: no cover
    return Markup(result)



import argparse
import sys
from typing import List, NoReturn, Optional

from strictdoc import __version__
from strictdoc.backend.reqif.sdoc_reqif_fields import ReqIFProfile
from strictdoc.backend.sdoc.constants import SDocMarkup
from strictdoc.cli.argument_int_range import IntRange

EXPORT_FORMATS = [
    "html",
    "html2pdf",
    "rst",
    "json",
    "excel",
    "reqif-sdoc",
    "reqifz-sdoc",
    "sdoc",
    "doxygen",
    "spdx",
]
EXCEL_PARSERS = ["basic"]


def formatter(prog: str) -> argparse.RawTextHelpFormatter:
    return argparse.RawTextHelpFormatter(
        prog, indent_increment=2, max_help_position=4, width=80
    )


def _check_formats(formats: str) -> List[str]:
    formats_array = formats.split(",")
    for fmt in formats_array:
        if fmt in EXPORT_FORMATS:
            continue
        export_formats = ", ".join(map(lambda f: f"'{f}'", EXPORT_FORMATS))
        message = f"invalid choice: '{fmt}' (choose from {export_formats})"
        raise argparse.ArgumentTypeError(message)
    return formats_array


def _check_reqif_profile(profile: str) -> str:
    if profile not in ReqIFProfile.ALL:
        # To maintain the compatibility with the previous behavior.
        if profile == "sdoc":
            return ReqIFProfile.P01_SDOC
        valid_profiles = ", ".join(map(lambda f: f"'{f}'", ReqIFProfile.ALL))
        message = f"invalid choice: '{profile}' (choose from {valid_profiles})"
        raise argparse.ArgumentTypeError(message)
    return profile


def _check_reqif_import_markup(markup: Optional[str]) -> str:
    if markup is None or markup not in SDocMarkup.ALL:
        valid_text_markups_string = ", ".join(SDocMarkup.ALL)
        message = f"invalid choice: '{markup}' (choose from {valid_text_markups_string})"
        raise argparse.ArgumentTypeError(message)
    return markup


def _parse_fields(fields: str) -> List[str]:
    fields_array = fields.split(",")
    return fields_array


def add_config_argument(parser: argparse.ArgumentParser) -> None:
    parser.add_argument(
        "--config",
        type=str,
        help="Path to the StrictDoc TOML config file.",
    )


class SDocArgumentParser(argparse.ArgumentParser):
    def error(self, message: str) -> NoReturn:
        self.print_usage(sys.stderr)
        print(f"{self.prog}: error: {message}", file=sys.stderr)  # noqa: T201
        print("")  # noqa: T201
        print("Further help:")  # noqa: T201
        print(  # noqa: T201
            "'strictdoc -h/--help' provides a general overview of available commands."
        )
        print(  # noqa: T201
            "'strictdoc <command> -h/--help' provides command-specific help."
        )
        sys.exit(2)


class CommandParserBuilder:
    def build(self) -> SDocArgumentParser:
        # https://stackoverflow.com/a/19476216/598057
        main_parser = SDocArgumentParser(
            prog="strictdoc",
            add_help=True,
            epilog=(
                """
                Further help: https://strictdoc.readthedocs.io/en/stable/
                """
            ),
        )

        # The -v/--version has a special behavior that it still works when all
        # commands are required == True.
        # https://stackoverflow.com/a/12123598/598057
        main_parser.add_argument(
            "-v", "--version", action="version", version=__version__
        )

        main_parser.add_argument(
            "--debug",
            action="store_true",
            default=False,
            help="Enable more verbose printing of errors when they are encountered.",
        )

        command_subparsers = main_parser.add_subparsers(
            title="command", dest="command"
        )
        command_subparsers.required = True

        self.add_about_command(command_subparsers)
        self.add_export_command(command_subparsers)
        self.add_server_command(command_subparsers)
        self.add_manage_command(command_subparsers)
        self.add_import_command(command_subparsers)
        self.add_version_command(command_subparsers)
        self.add_passthrough_command(command_subparsers)
        self.add_dump_command(command_subparsers)
        self.add_diff_command(command_subparsers)

        return main_parser

    @staticmethod
    def add_about_command(
        parent_command_parser: "argparse._SubParsersAction[SDocArgumentParser]",
    ) -> None:
        parent_command_parser.add_parser(
            "about",
            help="About StrictDoc.",
            description="About StrictDoc.",
            formatter_class=formatter,
        )

    @staticmethod
    def add_export_command(
        parent_command_parser: "argparse._SubParsersAction[SDocArgumentParser]",
    ) -> None:
        # Command – Export.
        command_parser_export = parent_command_parser.add_parser(
            "export",
            help="Export document tree.",
            description=(
                "Export command: "
                "input SDoc documents are generated into "
                "HTML and other formats."
            ),
            formatter_class=formatter,
        )
        command_parser_export.add_argument(
            "input_paths",
            type=str,
            nargs="+",
            help="One or more folders with *.sdoc files",
        )
        command_parser_export.add_argument(
            "--output-dir", type=str, help="Output folder"
        )
        command_parser_export.add_argument(
            "--project-title", type=str, help="Project title"
        )
        command_parser_export.add_argument(
            "--formats",
            type=_check_formats,
            default=["html"],
            help="Export formats",
        )
        command_parser_export.add_argument(
            "--fields",
            type=_parse_fields,
            default=["uid", "statement", "parent"],
            help="Export fields, only used for Excel export",
        )
        command_parser_export.add_argument(
            "--generate-bundle-document",
            action="store_true",
            default=False,
            help=(
                "EXPERIMENTAL: "
                "In addition to generating individual documents, "
                "also create a concatenated bundle that contains "
                "all the documents together."
            ),
        )
        command_parser_export.add_argument(
            "--no-parallelization",
            action="store_true",
            help=(
                "Disables parallelization. "
                "All work happens in the main thread. "
                "This option may be useful for debugging."
            ),
        )
        command_parser_export.add_argument(
            "--enable-mathjax",
            action="store_true",
            help="Enables Mathjax support (only HTML export).",
        )
        command_parser_export.add_argument(
            "--included-documents",
            action="store_true",
            help=(
                "By default the included documents are not exported. "
                "This option makes both including and included documents to be exported."
            ),
        )
        command_parser_export.add_argument(
            "--reqif-profile",
            type=_check_reqif_profile,
            default=ReqIFProfile.P01_SDOC,
            help="Export formats",
        )
        command_parser_export.add_argument(
            "--reqif-multiline-is-xhtml",
            default=False,
            action="store_true",
            help=(
                "Controls whether StrictDoc multiline fields are exported as XHTML "
                "when the option is provided. "
                "By default StrictDoc exports multiline fields with a STRING type."
            ),
        )
        command_parser_export.add_argument(
            "--reqif-enable-mid",
            default=False,
            action="store_true",
            help=(
                "Controls whether StrictDoc's MID field will be mapped to ReqIF "
                "SPEC-OBJECT's IDENTIFIER and vice versa when exporting/importing."
            ),
        )
        command_parser_export.add_argument(
            "--filter-requirements",
            type=str,
            help="Filter which requirements will be exported.",
        )
        command_parser_export.add_argument(
            "--filter-sections",
            type=str,
            help="Filter which sections will be exported.",
        )
        command_parser_export.add_argument(
            "--view",
            type=str,
            help="Choose which view will be exported.",
        )
        command_parser_export.add_argument(
            "--chromedriver",
            type=str,
            help="Path to pre installed chromedriver for html2pdf. "
            "If not given, chromedriver is downloaded and saved to "
            "strictdoc cache.",
        )
        add_config_argument(command_parser_export)

    @staticmethod
    def add_import_command(
        parent_command_parser: "argparse._SubParsersAction[SDocArgumentParser]",
    ) -> None:
        command_parser_import = parent_command_parser.add_parser(
            "import",
            help="Create StrictDoc files from other formats.",
            description="Create StrictDoc files from other formats.",
            formatter_class=formatter,
        )
        command_parser_import_subparsers = command_parser_import.add_subparsers(
            title="import_format", dest="import_format"
        )
        command_parser_import_subparsers.required = True

        # Command: Import -> ReqIF.
        command_parser_import_reqif = (
            command_parser_import_subparsers.add_parser(
                "reqif",
                help="Create StrictDoc file from ReqIF document.",
                description="Create StrictDoc file from ReqIF document.",
                formatter_class=formatter,
            )
        )

        command_parser_import_reqif.add_argument(
            "profile",
            type=_check_reqif_profile,
            help=(
                "An argument that selects the ReqIF import/export profile. "
                f"Possible values: {{{', '.join(ReqIFProfile.ALL)}}}"
            ),
        )
        command_parser_import_reqif.add_argument(
            "input_path",
            type=str,
            help="Path to the input ReqIF file.",
        )
        command_parser_import_reqif.add_argument(
            "output_path",
            type=str,
            help="Path to the output SDoc file.",
        )
        command_parser_import_reqif.add_argument(
            "--reqif-enable-mid",
            default=False,
            action="store_true",
            help=(
                "Controls whether StrictDoc's MID field will be mapped to ReqIF "
                "SPEC-OBJECT's IDENTIFIER and vice versa when exporting/importing."
            ),
        )
        command_parser_import_reqif.add_argument(
            "--reqif-import-markup",
            default=None,
            type=_check_reqif_import_markup,
            help=(
                "Controls which MARKUP option the imported SDoc documents will have. "
                "This value is RST as what StrictDoc has by default but very often "
                "the requirements tools use the (X)HTML markup for multiline fields in "
                "which case HTML is the best option."
            ),
        )

        # Command: Import -> Excel.
        command_parser_import_excel = (
            command_parser_import_subparsers.add_parser(
                "excel",
                help="Create StrictDoc file from Excel document.",
                description="Create StrictDoc file Excel ReqIF document.",
                formatter_class=formatter,
            )
        )

        def check_excel_parser(parser: str) -> str:
            if parser not in EXCEL_PARSERS:
                message = (
                    f"invalid choice: '{parser}' (choose from {EXCEL_PARSERS})"
                )
                raise argparse.ArgumentTypeError(message)
            return parser

        command_parser_import_excel.add_argument(
            "parser",
            type=check_excel_parser,
            help=(
                "An argument that selects the ReqIF parser. "
                f"Possible values: {{{', '.join(EXCEL_PARSERS)}}}"
            ),
        )
        command_parser_import_excel.add_argument(
            "input_path",
            type=str,
            help="Path to the input ReqIF file.",
        )
        command_parser_import_excel.add_argument(
            "output_path",
            type=str,
            help="Path to the output SDoc file.",
        )

    @staticmethod
    def add_server_command(
        parent_command_parser: "argparse._SubParsersAction[SDocArgumentParser]",
    ) -> None:
        command_parser_server = parent_command_parser.add_parser(
            "server",
            help="Run StrictDoc web server.",
            description="Run StrictDoc web server.",
            formatter_class=formatter,
        )
        command_parser_server.add_argument("input_path")
        command_parser_server.add_argument("--output-path", type=str)
        command_parser_server.add_argument("--host", type=str)
        command_parser_server.add_argument("--port", type=IntRange(1024, 65000))
        # The --reload and --no-reload options are currently used only for
        # StrictDoc's own development. We may want to revisit this for the end
        # users in the future but for now they are excluded from the help
        # messages.
        command_parser_server.add_argument(
            "--reload",
            default=False,
            action="store_true",
            help=argparse.SUPPRESS,
        )
        command_parser_server.add_argument(
            "--no-reload",
            dest="reload",
            action="store_false",
            help=argparse.SUPPRESS,
        )

        add_config_argument(command_parser_server)

    @staticmethod
    def add_version_command(
        parent_command_parser: "argparse._SubParsersAction[SDocArgumentParser]",
    ) -> None:
        parent_command_parser.add_parser(
            "version",
            help="Print the version of StrictDoc.",
            formatter_class=formatter,
        )

    @staticmethod
    def add_passthrough_command(
        parent_command_parser: "argparse._SubParsersAction[SDocArgumentParser]",
    ) -> None:
        # Passthrough command is kept for backwards compatibility. It will
        # internally be handled as export --formats sdoc.
        command_parser_passthrough = parent_command_parser.add_parser(
            "passthrough",
            help="Read an SDoc file, then output it again. Deprecated, use "
            "strictdoc export --formats sdoc instead.",
            formatter_class=formatter,
        )
        command_parser_passthrough.add_argument(
            "input_paths",
            type=str,
            nargs="+",
            help="One or more folders with *.sdoc files",
        )
        command_parser_passthrough.add_argument(
            "--output-dir",
            type=str,
            help="A directory where to output the SDoc files to.",
        )
        command_parser_passthrough.add_argument(
            "--filter-requirements",
            type=str,
            help="Filter which requirements will be exported.",
        )
        command_parser_passthrough.add_argument(
            "--filter-sections",
            type=str,
            help="Filter which sections will be exported.",
        )
        command_parser_passthrough.add_argument(
            "--view",
            type=str,
            help="Choose which view will be exported.",
        )

        # Hidden default values to make deprecated passthrough parser compatible
        # with SDocArgsParser.get_export_config.
        command_parser_passthrough.add_argument(
            "--project-title", const=None, help=argparse.SUPPRESS
        )
        command_parser_passthrough.add_argument(
            "--formats",
            default=["sdoc"],
            help=argparse.SUPPRESS,
        )
        command_parser_passthrough.add_argument(
            "--fields",
            default=None,
            help=argparse.SUPPRESS,
        )
        command_parser_passthrough.add_argument(
            "--generate-bundle-document",
            default=False,
            help=argparse.SUPPRESS,
        )
        command_parser_passthrough.add_argument(
            "--no-parallelization",
            default=False,
            help=argparse.SUPPRESS,
        )
        command_parser_passthrough.add_argument(
            "--enable-mathjax",
            default=False,
            help=argparse.SUPPRESS,
        )
        command_parser_passthrough.add_argument(
            "--included-documents",
            default=False,
            help=argparse.SUPPRESS,
        )
        command_parser_passthrough.add_argument(
            "--reqif-profile",
            default=None,
            help=argparse.SUPPRESS,
        )
        command_parser_passthrough.add_argument(
            "--reqif-multiline-is-xhtml",
            default=False,
            help=argparse.SUPPRESS,
        )
        command_parser_passthrough.add_argument(
            "--reqif-enable-mid",
            default=False,
            help=argparse.SUPPRESS,
        )
        command_parser_passthrough.add_argument(
            "--chromedriver",
            default=None,
            help=argparse.SUPPRESS,
        )
        add_config_argument(command_parser_passthrough)

    @staticmethod
    def add_dump_command(
        parent_command_parser: "argparse._SubParsersAction[SDocArgumentParser]",
    ) -> None:
        command_parser_dump_grammar = parent_command_parser.add_parser(
            "dump-grammar",
            help="Dump the SDoc grammar to a .tx file.",
            formatter_class=formatter,
        )
        command_parser_dump_grammar.add_argument(
            "output_file", type=str, help="Path to the output .tx file"
        )

    @staticmethod
    def add_manage_command(
        parent_command_parser: "argparse._SubParsersAction[SDocArgumentParser]",
    ) -> None:
        manage_command_parser = parent_command_parser.add_parser(
            "manage",
            help="Manage StrictDoc project.",
            description="See subcommands to manage StrictDoc project.",
            formatter_class=formatter,
        )
        manage_command_subparsers = manage_command_parser.add_subparsers(
            title="subcommand", dest="subcommand"
        )
        manage_command_subparsers.required = True

        command_parser_auto_uid = manage_command_subparsers.add_parser(
            "auto-uid",
            help="Generates missing requirements UIDs automatically.",
            description=(
                "This command generates missing requirement UID automatically. "
                "The UIDs are generated based on the nearest section "
                "PREFIX (if provided) or the document's "
                'PREFIX (if provided or "REQ-" by default).'
            ),
            formatter_class=formatter,
        )

        command_parser_auto_uid.add_argument(
            "input_path",
            type=str,
            help="Path to the project tree.",
        )
        command_parser_auto_uid.add_argument(
            "--include-sections",
            action="store_true",
            help=(
                "By default, the command only generates the UID for "
                "requirements. This option enables the generation of UID for "
                "sections."
            ),
        )
        add_config_argument(command_parser_auto_uid)

    @staticmethod
    def add_diff_command(
        command_subparsers: "argparse._SubParsersAction[SDocArgumentParser]",
    ) -> None:
        diff_command_parser = command_subparsers.add_parser(
            "diff",
            help="Generate Diff between two SDoc project trees.",
            formatter_class=formatter,
        )
        diff_command_parser.add_argument(
            "path_to_lhs_tree",
            type=str,
            help="Path to the left-hand side project tree.",
        )
        diff_command_parser.add_argument(
            "path_to_rhs_tree",
            type=str,
            help="Path to the right-hand side project tree.",
        )
        diff_command_parser.add_argument(
            "--output-dir",
            type=str,
            help="A directory where to output the files to.",
        )



"""
Class container that stores document meta information.

Example explaining metadata stored by this class:.

DocumentMeta(
    level = 1,
    file_tree_mount_folder = "doc_project",
    document_filename = "sample.sdoc",
    document_filename_base = "sample",
    input_doc_full_path = "/tmp/doc_project/child.sdoc",
    input_doc_rel_path = "child.sdoc",
    input_doc_dir_rel_path = "",
    input_doc_assets_dir_rel_path = "doc_project/_assets",
    output_document_dir_full_path = "/tmp/doc_project/output/html/doc_project",
    output_document_dir_rel_path = "doc_project"
)
"""

import os
from typing import Optional

from strictdoc.export.html.document_type import DocumentType
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.paths import SDocRelativePath


@auto_described
class DocumentMeta:
    def __init__(
        self,
        level: int,
        file_tree_mount_folder: str,
        document_filename: str,
        document_filename_base: str,
        input_doc_full_path: str,
        input_doc_rel_path: SDocRelativePath,
        input_doc_dir_rel_path: SDocRelativePath,
        input_doc_assets_dir_rel_path: SDocRelativePath,
        output_document_dir_full_path: str,
        output_document_dir_rel_path: SDocRelativePath,
    ) -> None:
        assert isinstance(input_doc_rel_path, SDocRelativePath), (
            input_doc_rel_path
        )
        assert isinstance(input_doc_dir_rel_path, SDocRelativePath), (
            input_doc_dir_rel_path
        )
        assert isinstance(input_doc_assets_dir_rel_path, SDocRelativePath), (
            input_doc_assets_dir_rel_path
        )
        assert isinstance(output_document_dir_rel_path, SDocRelativePath), (
            output_document_dir_rel_path
        )

        self.level: int = level
        self.file_tree_mount_folder = file_tree_mount_folder
        self.document_filename: str = document_filename
        self.document_filename_base: str = document_filename_base
        self.input_doc_full_path: str = input_doc_full_path
        self.input_doc_rel_path: SDocRelativePath = input_doc_rel_path
        self.input_doc_dir_rel_path: SDocRelativePath = input_doc_dir_rel_path
        self.input_doc_assets_dir_rel_path: SDocRelativePath = (
            input_doc_assets_dir_rel_path
        )
        self.output_document_dir_full_path: str = output_document_dir_full_path
        self.output_document_dir_rel_path: SDocRelativePath = (
            output_document_dir_rel_path
        )
        self.output_document_full_path: str = os.path.join(
            self.output_document_dir_full_path,
            self.get_html_doc_path(),
        )

    def get_html_doc_path(self) -> str:
        return (
            f"{self.output_document_dir_full_path}"
            f"/"
            f"{self.document_filename_base}.html"
        )

    def get_html_doc_standalone_path(self) -> str:
        return (
            f"{self.output_document_dir_full_path}"
            f"/"
            f"{self.document_filename_base}.standalone.html"
        )

    def get_html_table_path(self) -> str:
        return (
            f"{self.output_document_dir_full_path}"
            f"/"
            f"{self.document_filename_base}-TABLE.html"
        )

    def get_html_traceability_path(self) -> str:
        return (
            f"{self.output_document_dir_full_path}"
            f"/"
            f"{self.document_filename_base}-TRACE.html"
        )

    def get_html_deep_traceability_path(self) -> str:
        return (
            f"{self.output_document_dir_full_path}"
            f"/"
            f"{self.document_filename_base}-DEEP-TRACE.html"
        )

    def get_html_pdf_path(self) -> str:
        return (
            f"{self.output_document_dir_full_path}"
            f"/"
            f"{self.document_filename_base}-PDF.html"
        )

    # Links
    def get_html_doc_link(self) -> str:
        file_name_part = f"{self.document_filename_base}.html"
        if len(self.output_document_dir_rel_path.relative_path_posix) == 0:
            return file_name_part
        return (
            f"{self.output_document_dir_rel_path.relative_path_posix}"
            "/"
            f"{file_name_part}"
        )

    def get_html_table_link(self) -> str:
        file_name_part = f"{self.document_filename_base}-TABLE.html"
        if len(self.output_document_dir_rel_path.relative_path_posix) == 0:
            return file_name_part
        return (
            f"{self.output_document_dir_rel_path.relative_path_posix}"
            "/"
            f"{file_name_part}"
        )

    def get_html_traceability_link(self) -> str:
        file_name_part = f"{self.document_filename_base}-TRACE.html"
        if len(self.output_document_dir_rel_path.relative_path_posix) == 0:
            return file_name_part
        return (
            f"{self.output_document_dir_rel_path.relative_path_posix}"
            "/"
            f"{file_name_part}"
        )

    def get_html_deep_traceability_link(self) -> str:
        file_name_part = f"{self.document_filename_base}-DEEP-TRACE.html"
        if len(self.output_document_dir_rel_path.relative_path_posix) == 0:
            return file_name_part
        return (
            f"{self.output_document_dir_rel_path.relative_path_posix}"
            "/"
            f"{file_name_part}"
        )

    def get_html_pdf_link(self) -> str:
        file_name_part = f"{self.document_filename_base}-PDF.html"
        if len(self.output_document_dir_rel_path.relative_path_posix) == 0:
            return file_name_part
        return (
            f"{self.output_document_dir_rel_path.relative_path_posix}"
            "/"
            f"{file_name_part}"
        )

    def get_html_standalone_document_link(self) -> str:
        return (
            f"{self.output_document_dir_rel_path.relative_path_posix}"
            f"/"
            f"{self.document_filename_base}.standalone.html"
        )

    def get_html_link(
        self,
        document_type: DocumentType,
        other_doc_level: Optional[int],
    ) -> str:
        assert isinstance(document_type, DocumentType)

        path_prefix: str = self.get_root_path_prefix(other_doc_level)
        if document_type == DocumentType.DOCUMENT:
            document_link = self.get_html_doc_link()
        elif document_type == DocumentType.TABLE:
            document_link = self.get_html_table_link()
        elif document_type == DocumentType.TRACE:
            document_link = self.get_html_traceability_link()
        elif document_type == DocumentType.DEEPTRACE:
            document_link = self.get_html_deep_traceability_link()
        elif document_type == DocumentType.PDF:
            document_link = self.get_html_pdf_link()
        else:
            raise NotImplementedError
        # We reach here the document is a bundle document.
        if len(path_prefix) == 0:
            return document_link
        return f"{path_prefix}/{document_link}"

    def get_root_path_prefix(
        self, other_doc_level: Optional[int] = None
    ) -> str:
        level: int = self.level if not other_doc_level else other_doc_level
        if level == 0:
            return ""
        return ("../" * level)[:-1]



import typing
from collections import Counter
from typing import Dict, List

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.analyzers.document_stats import (
    DocumentStats,
    DocumentTreeStats,
    SinglePrefixRequirements,
)
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.string import (
    extract_last_numeric_part,
    extract_numeric_uid_prefix_part,
)


class DocumentUIDAnalyzer:
    @staticmethod
    def analyze_document_tree(
        traceability_index: TraceabilityIndex,
    ) -> DocumentTreeStats:
        global_requirements_per_prefix: Dict[str, SinglePrefixRequirements] = {}
        document_tree_stats: List[DocumentStats] = []
        section_uids_so_far: typing.Counter[str] = Counter()

        document_tree: DocumentTree = assert_cast(
            traceability_index.document_tree, DocumentTree
        )

        for document in document_tree.document_list:
            document_stats: DocumentStats = (
                DocumentUIDAnalyzer.analyze_document(document)
            )
            for (
                requirement_prefix_,
                prefix_requirements_,
            ) in document_stats.requirements_per_prefix.items():
                global_prefix_requirements = (
                    global_requirements_per_prefix.setdefault(
                        requirement_prefix_, SinglePrefixRequirements()
                    )
                )
                global_prefix_requirements.requirements_no_uid.extend(
                    prefix_requirements_.requirements_no_uid
                )
                global_prefix_requirements.requirements_uid_numbers.extend(
                    prefix_requirements_.requirements_uid_numbers
                )
            document_tree_stats.append(document_stats)
            for section_uid_ in document_stats.section_uids_so_far:
                section_uids_so_far[section_uid_] += 1
        return DocumentTreeStats(
            single_document_stats=document_tree_stats,
            requirements_per_prefix=global_requirements_per_prefix,
            section_uids_so_far=section_uids_so_far,
        )

    @staticmethod
    def analyze_document(
        document: SDocDocument,
    ) -> DocumentStats:
        this_document_stats = DocumentStats(document)
        document_iterator = DocumentCachingIterator(document)
        for node, _ in document_iterator.all_content():
            if isinstance(node, SDocSection) or (
                isinstance(node, SDocNode) and node.node_type == "SECTION"
            ):
                if node.reserved_uid is not None:
                    this_document_stats.section_uids_so_far[
                        node.reserved_uid
                    ] += 1
                else:
                    this_document_stats.sections_without_uid.append(node)
                continue
            if not isinstance(node, SDocNode):
                continue
            if node.node_type in ("TEXT", "SECTION"):
                continue
            requirement: SDocNode = node
            node_prefix: typing.Optional[str] = requirement.get_prefix()
            if node_prefix is None:
                continue

            if node_prefix not in this_document_stats.requirements_per_prefix:
                this_document_stats.requirements_per_prefix[node_prefix] = (
                    SinglePrefixRequirements()
                )

            prefix_requirements: SinglePrefixRequirements = (
                this_document_stats.requirements_per_prefix[node_prefix]
            )
            if requirement.reserved_uid is None:
                prefix_requirements.requirements_no_uid.append(requirement)
            else:
                requirement_prefix = extract_numeric_uid_prefix_part(
                    requirement.reserved_uid
                )

                # If the requirement has a prefix which is different to that of
                # the parent section/document's prefix, we ignore this node as
                # incredible for contributing to the next UID calculation.
                if requirement_prefix != node_prefix:
                    continue

                requirement_uid_numeric_part = extract_last_numeric_part(
                    requirement.reserved_uid
                )
                assert len(requirement_uid_numeric_part) > 0
                requirement_uid_number = int(requirement_uid_numeric_part)

                prefix_requirements.requirements_uid_numbers.append(
                    requirement_uid_number
                )

        return this_document_stats



import os
from typing import List

from docutils import nodes
from docutils.parsers.rst.directives.images import Image


class WildcardEnhancedImage(Image):  # type: ignore[misc]
    WILDCARD_EXTENSIONS = ["svg", "png", "gif", "jpg", "jpeg"]

    current_reference_path = os.getcwd()

    def run(self) -> List[nodes.Node]:
        # A user has suggested that StrictDoc could be capable of rendering
        # a Sphinx-specific directive: image.* (this directive does not exist
        # in the Docutils).
        # This directive allows a specification of "image.*", and the wildcard
        # extension is replaced by a search over a number of extensions.
        # We want StrictDoc to recognize this custom .* prefix and interpret
        # it with a search over a number of extensions. If a file with a given
        # extension is found, we pass it to the original Image directive, and
        # it works as normal.
        #
        # User request:
        # https://github.com/strictdoc-project/strictdoc/issues/1106
        # The Sphinx directive is documented here:
        # https://www.sphinx-doc.org/en/master/usage/restructuredtext/basics.html#images
        #
        # Sphinx/Docutils files with relevant implementation:
        # - docutils/docutils/parsers/rst/directives/images.py
        # - sphinx/writers/html5.py (visit_image)  # noqa: ERA001
        #
        # When this method is called from the RST string:
        # """
        # .. image:: some_picture.*
        # """,
        # the self.arguments looks like this:
        # ['some_picture.*']  # noqa: ERA001
        messages: List[nodes.Node] = []

        assert len(self.arguments) > 0
        rel_path_to_image = self.arguments[0]
        if rel_path_to_image.endswith(".*"):
            rel_path_to_image_no_wc = rel_path_to_image[:-2]
            for extension in WildcardEnhancedImage.WILDCARD_EXTENSIONS:
                rel_path_to_image_with_extension = (
                    rel_path_to_image_no_wc + "." + extension
                )
                full_path_to_image_with_extension = os.path.join(
                    WildcardEnhancedImage.current_reference_path,
                    rel_path_to_image_with_extension,
                )
                if os.path.exists(full_path_to_image_with_extension):
                    # We have found a matching file, let's use it.
                    self.arguments[0] = os.path.join(
                        rel_path_to_image_with_extension
                    )
                    break
            else:
                # If the argument is not provided, raise an error.
                error_message = (
                    f"No image could be found to match the wildcard: "
                    f"{rel_path_to_image}"
                )
                self.state_machine.reporter.error(
                    error_message, line=self.lineno
                )
                # Return an empty list of nodes to stop the rendering
                # process.
                return []

        messages = super().run()

        return messages



from typing import Any, Dict, Hashable, List, Optional, Tuple

from strictdoc.core.graph.abstract_bucket import AbstractBucket
from strictdoc.helpers.ordered_set import OrderedSet


class ConstraintViolation(Exception):
    def __init__(self, message: str) -> None:
        super().__init__(message)


class GraphDatabase:
    def __init__(self, buckets: List[Tuple[Any, AbstractBucket]]):
        self._links = None
        self._id_to_bucket: Dict[Any, AbstractBucket] = {}
        for bucket_ in buckets:
            assert bucket_[0] not in self._id_to_bucket
            self._id_to_bucket[bucket_[0]] = bucket_[1]

    def has_link(self, *, link_type: Hashable, lhs_node: Any) -> bool:
        return self._id_to_bucket[link_type].has_link(lhs_node=lhs_node)

    def get_count(self, *, link_type: Hashable) -> int:
        return self._id_to_bucket[link_type].get_count()

    def get_link_value_weak(
        self, *, link_type: Hashable, lhs_node: Any
    ) -> Optional[Any]:
        return self._id_to_bucket[link_type].get_link_value_weak(
            lhs_node=lhs_node
        )

    def get_link_value(
        self, *, link_type: Hashable, lhs_node: Any, edge: Optional[str] = None
    ) -> Any:
        return self._id_to_bucket[link_type].get_link_value(
            lhs_node=lhs_node, edge=edge
        )

    def get_link_values(
        self, *, link_type: Hashable, lhs_node: Any, edge: Optional[str] = None
    ) -> OrderedSet[Any]:
        return self._id_to_bucket[link_type].get_link_values(
            lhs_node=lhs_node, edge=edge
        )

    def get_link_values_with_edges(
        self, *, link_type: Hashable, lhs_node: Any, edge: Optional[str] = None
    ) -> List[Tuple[Any, Optional[str]]]:
        return self._id_to_bucket[link_type].get_link_values_with_edges(
            lhs_node=lhs_node, edge=edge
        )

    def get_link_values_reverse(
        self, *, link_type: Hashable, rhs_node: Any
    ) -> Any:
        return self._id_to_bucket[link_type].get_link_values_reverse(
            rhs_node=rhs_node
        )

    def create_link(
        self,
        *,
        link_type: Hashable,
        lhs_node: Any,
        rhs_node: Any,
        edge: Optional[str] = None,
    ) -> None:
        assert lhs_node is not None, lhs_node
        assert rhs_node is not None, rhs_node
        self._id_to_bucket[link_type].create_link(
            lhs_node=lhs_node, rhs_node=rhs_node, edge=edge
        )

    def delete_link(
        self,
        *,
        link_type: Hashable,
        lhs_node: Any,
        rhs_node: Any,
        edge: Optional[str] = None,
    ) -> None:
        self._id_to_bucket[link_type].delete_link(
            lhs_node=lhs_node, rhs_node=rhs_node, edge=edge
        )

    def delete_link_weak(
        self,
        *,
        link_type: Hashable,
        lhs_node: Any,
        rhs_node: Any,
    ) -> None:
        self._id_to_bucket[link_type].delete_link_weak(
            lhs_node=lhs_node, rhs_node=rhs_node
        )

    def delete_all_links(
        self,
        *,
        link_type: Hashable,
        lhs_node: Any,
    ) -> None:
        self._id_to_bucket[link_type].delete_all_links(lhs_node=lhs_node)



import hashlib
import os
import stat

from starlette.requests import Request


def get_etag(file_path: str) -> str:
    """
    Get an HTTP ETag value for a given file path.

    This implementation is taken from StackOverflow. The implementation itself
    is taken from Starlette.
    https://stackoverflow.com/a/76263874/598057

    NOTE: The Starlette's version is async and uses anyio.
    """

    assert os.path.isfile(file_path)
    stat_result = os.stat(file_path)
    mode = stat_result.st_mode
    if not stat.S_ISREG(mode):
        raise RuntimeError(f"File at path {file_path} is not a file.")

    # Calculate the etag based on file size and last modification time.
    etag_base = str(stat_result.st_mtime) + "-" + str(stat_result.st_size)
    etag = hashlib.md5(etag_base.encode()).hexdigest()
    return etag


def request_is_for_non_modified_file(
    request: Request, static_file: str
) -> bool:
    if "if-none-match" not in request.headers:
        return False

    # This request header comes up with quotes. Strip them.
    header_etag: str = request.headers["if-none-match"].strip('"')

    # We have copied the Etag calculation procedure from
    # Starlette's server code. One day this copy may diverge if
    # Starlette decides to implement something else.
    # In that case, the risk is that the 200/304 caching will stop
    # working but such a risk seems acceptable. There is an e2e test
    # that verifies that this branch is reached.
    file_etag = get_etag(static_file)
    return header_etag == file_etag



from markupsafe import Markup, escape


class TextToHtmlWriter:
    @staticmethod
    def write(text_fragment: str) -> Markup:
        return escape(text_fragment).replace("\n", Markup("<br/>\n"))

    @staticmethod
    def write_anchor_link(title: str, href: str) -> str:
        return f'<a href="{href}">🔗&nbsp;{title}</a>`'



from enum import Enum
from typing import Tuple


class DocumentType(str, Enum):
    DOCUMENT = "DOCUMENT"
    TABLE = "TABLE"
    TRACE = "TRACE"
    DEEPTRACE = "DEEPTRACE"
    PDF = "PDF"

    @staticmethod
    def all() -> Tuple["DocumentType", ...]:  # noqa: A003
        return (
            DocumentType.DOCUMENT,
            DocumentType.TABLE,
            DocumentType.TRACE,
            DocumentType.DEEPTRACE,
            DocumentType.PDF,
        )

    def is_document(self) -> bool:
        return self == DocumentType.DOCUMENT

    def is_table(self) -> bool:
        return self == DocumentType.TABLE

    def is_trace(self) -> bool:
        return self == DocumentType.TRACE

    def is_deeptrace(self) -> bool:
        return self == DocumentType.DEEPTRACE

    def is_pdf(self) -> bool:
        return self == DocumentType.PDF

    def get_page_title(self) -> str:
        if self == DocumentType.DOCUMENT:
            return "Document"
        if self == DocumentType.TABLE:
            return "Table"
        if self == DocumentType.TRACE:
            return "Traceability"
        if self == DocumentType.DEEPTRACE:
            return "Deep Traceability"
        if self == DocumentType.PDF:
            return "PDF"
        raise NotImplementedError



from typing import Dict, List, Optional

from starlette.datastructures import FormData

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.form_data import ParsedFormData, parse_form_data
from strictdoc.helpers.mid import MID
from strictdoc.helpers.string import sanitize_html_form_field
from strictdoc.server.error_object import ErrorObject


@auto_described
class DocumentMetadataFormField:
    def __init__(
        self,
        field_mid: str,
        field_name: str,
        field_value: str,
    ):
        assert isinstance(field_value, str)
        self.field_mid: str = field_mid
        self.field_name: str = field_name
        self.field_value: str = field_value

    @staticmethod
    def create_from_document(
        name: str, value: str
    ) -> "DocumentMetadataFormField":
        return DocumentMetadataFormField(
            field_mid=MID.create(), field_name=name, field_value=value
        )

    def get_input_field_name(self) -> str:
        return f"metadata[{self.field_mid}][name]"

    def get_input_field_value(self) -> str:
        return f"metadata[{self.field_mid}][value]"


@auto_described
class DocumentConfigFormObject(ErrorObject):
    def __init__(
        self,
        *,
        document_mid: str,
        document_title: str,
        document_uid: Optional[str],
        document_version: Optional[str],
        document_classification: Optional[str],
        document_requirement_prefix: Optional[str],
        document_custom_metadata_fields: List[DocumentMetadataFormField],
    ):
        assert isinstance(document_mid, str), document_mid
        assert isinstance(document_title, str), document_title
        super().__init__()
        self.document_mid: Optional[str] = document_mid
        self.document_title: str = document_title
        self.document_uid: Optional[str] = (
            document_uid if document_uid is not None else ""
        )
        self.document_version: Optional[str] = (
            document_version if document_version is not None else ""
        )
        self.document_classification: Optional[str] = (
            document_classification
            if document_classification is not None
            else ""
        )
        self.document_requirement_prefix: Optional[str] = (
            document_requirement_prefix
            if document_requirement_prefix is not None
            else ""
        )
        self.custom_metadata_fields: List[DocumentMetadataFormField] = (
            document_custom_metadata_fields
        )

    @staticmethod
    def create_from_request(
        *, document_mid: str, request_form_data: FormData
    ) -> "DocumentConfigFormObject":
        request_form_data_as_list = [
            (field_name, field_value)
            for field_name, field_value in request_form_data.multi_items()
        ]
        request_form_dict: ParsedFormData = assert_cast(
            parse_form_data(request_form_data_as_list), dict
        )

        config_fields: Dict[str, str] = assert_cast(
            request_form_dict["document"], dict
        )

        metadata_fields: Dict[str, Dict[str, str]] = (
            assert_cast(request_form_dict["metadata"], dict)
            if "metadata" in request_form_dict
            else {}
        )

        document_title: str = (
            config_fields["TITLE"] if "TITLE" in config_fields else ""
        )
        document_title = sanitize_html_form_field(
            document_title, multiline=False
        )
        document_title = document_title if document_title is not None else ""

        document_uid: str = ""
        if "UID" in config_fields:
            document_uid = config_fields["UID"]
            document_uid = sanitize_html_form_field(
                document_uid, multiline=False
            )

        document_version: str = ""
        if "VERSION" in config_fields:
            document_version = config_fields["VERSION"]
            document_version = sanitize_html_form_field(
                document_version, multiline=False
            )

        document_classification: str = ""
        if "CLASSIFICATION" in config_fields:
            document_classification = config_fields["CLASSIFICATION"]
            document_classification = sanitize_html_form_field(
                document_classification, multiline=False
            )

        document_requirement_prefix: str = ""
        if "PREFIX" in config_fields:
            document_requirement_prefix = config_fields["PREFIX"]
            document_requirement_prefix = sanitize_html_form_field(
                document_requirement_prefix, multiline=False
            )

        document_custom_metadata_fields: List[DocumentMetadataFormField] = []
        for field_mid, field_dict in metadata_fields.items():
            assert isinstance(field_dict, dict), type(field_dict)

            field_name = field_dict["name"].strip()
            field_value = field_dict["value"].strip()

            document_custom_metadata_field = DocumentMetadataFormField(
                field_mid=field_mid,
                field_name=field_name,
                field_value=field_value,
            )
            document_custom_metadata_fields.append(
                document_custom_metadata_field
            )

        form_object = DocumentConfigFormObject(
            document_mid=document_mid,
            document_title=document_title,
            document_uid=document_uid,
            document_version=document_version,
            document_classification=document_classification,
            document_requirement_prefix=document_requirement_prefix,
            document_custom_metadata_fields=document_custom_metadata_fields,
        )
        return form_object

    @staticmethod
    def create_from_document(
        *,
        document: SDocDocument,
    ) -> "DocumentConfigFormObject":
        assert isinstance(document, SDocDocument)

        document_custom_metadata_fields: List[DocumentMetadataFormField] = []
        for name, value in document.config.get_custom_metadata():
            document_custom_metadata_field = (
                DocumentMetadataFormField.create_from_document(
                    name=name, value=value
                )
            )
            document_custom_metadata_fields.append(
                document_custom_metadata_field
            )

        return DocumentConfigFormObject(
            document_mid=document.reserved_mid,
            document_title=document.title,
            document_uid=document.config.uid,
            document_version=document.config.version,
            document_classification=document.config.classification,
            document_requirement_prefix=document.config.requirement_prefix,
            document_custom_metadata_fields=document_custom_metadata_fields,
        )



from strictdoc.backend.sdoc.grammar.grammar_builder import SDocGrammarBuilder
from strictdoc.cli.cli_arg_parser import DumpGrammarCommandConfig


class DumpGrammarCommand:
    @staticmethod
    def execute(config: DumpGrammarCommandConfig) -> None:
        with open(config.output_file, "w", encoding="utf8") as output_file:
            output_file.write(SDocGrammarBuilder.create_raw_grammar())



import sys
import traceback
from typing import Optional

from textx import metamodel_from_str

from strictdoc.backend.sdoc.error_handling import StrictDocSemanticError
from strictdoc.backend.sdoc.grammar.grammar_builder import SDocGrammarBuilder
from strictdoc.backend.sdoc.models.constants import GRAMMAR_MODELS
from strictdoc.backend.sdoc.models.document_grammar import (
    DocumentGrammar,
    DocumentGrammarWrapper,
)
from strictdoc.backend.sdoc.pickle_cache import PickleCache
from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.cast import assert_optional_cast
from strictdoc.helpers.textx import (
    drop_textx_meta,
    preserve_source_location_data,
)


class SDocGrammarReader:
    meta_model = metamodel_from_str(
        SDocGrammarBuilder.create_grammar_grammar(),
        classes=GRAMMAR_MODELS + [DocumentGrammarWrapper],
        use_regexp_group=True,
    )

    @staticmethod
    def read(
        input_string: str, file_path: Optional[str] = None
    ) -> DocumentGrammar:
        SDocGrammarReader.meta_model.register_obj_processors(
            {
                "GrammarElement": preserve_source_location_data,
            }
        )

        grammar_wrapper: DocumentGrammarWrapper = (
            SDocGrammarReader.meta_model.model_from_str(
                input_string, file_name=file_path
            )
        )
        grammar: DocumentGrammar = grammar_wrapper.grammar
        grammar.parent = None

        # HACK:
        # ProcessPoolExecutor doesn't work because of non-picklable parts
        # of textx. The offending fields are stripped down because they
        # are not used anyway.
        drop_textx_meta(grammar)

        return grammar

    def read_from_file(
        self, file_path: str, project_config: ProjectConfig
    ) -> DocumentGrammar:
        unpickled_content: Optional[DocumentGrammar] = assert_optional_cast(
            PickleCache.read_from_cache(file_path, project_config, "grammar"),
            DocumentGrammar,
        )
        if unpickled_content is not None:
            return unpickled_content

        with open(file_path, encoding="utf8") as file:
            grammar_content = file.read()

        try:
            grammar: DocumentGrammar = self.read(
                grammar_content, file_path=file_path
            )
            PickleCache.save_to_cache(
                grammar, file_path, project_config, "grammar"
            )

            return grammar
        except NotImplementedError:
            traceback.print_exc()
            sys.exit(1)
        except StrictDocSemanticError as exc:
            print(exc.to_print_message())  # noqa: T201
            sys.exit(1)
        except Exception as exc:  # pylint: disable=broad-except
            print(  # noqa: T201
                f"error: could not parse file: "
                f"{file_path}.\n{exc.__class__.__name__}: {exc}"
            )
            # TODO: when --debug is provided.
            traceback.print_exc()
            sys.exit(1)



from abc import ABC, abstractmethod
from typing import Any, List, Optional, Tuple

from strictdoc.helpers.ordered_set import OrderedSet

ALL_EDGES = ".all"


class AbstractBucket(ABC):
    @abstractmethod
    def has_link(self, *, lhs_node: Any) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_count(self) -> int:
        raise NotImplementedError

    @abstractmethod
    def get_link_value(
        self, *, lhs_node: Any, edge: Optional[str] = None
    ) -> Any:
        raise NotImplementedError

    @abstractmethod
    def get_link_value_weak(self, *, lhs_node: Any) -> Optional[Any]:
        raise NotImplementedError

    def get_link_values(
        self, *, lhs_node: Any, edge: Optional[str] = None
    ) -> OrderedSet[Any]:
        raise NotImplementedError

    def get_link_values_with_edges(
        self, *, lhs_node: Any, edge: Optional[str] = None
    ) -> List[Tuple[Any, Optional[str]]]:
        raise NotImplementedError

    def get_link_values_reverse(self, *, rhs_node: Any) -> Any:
        raise NotImplementedError

    @abstractmethod
    def create_link(
        self, *, lhs_node: Any, rhs_node: Any, edge: Optional[str] = None
    ) -> None:
        raise NotImplementedError

    @abstractmethod
    def delete_link(
        self, *, lhs_node: Any, rhs_node: Any, edge: Optional[str] = None
    ) -> None:
        raise NotImplementedError

    def delete_link_weak(
        self,
        *,
        lhs_node: Any,
        rhs_node: Any,
    ) -> None:
        raise NotImplementedError

    def delete_all_links(
        self,
        *,
        lhs_node: Any,
    ) -> None:
        raise NotImplementedError



"""
@relation(SDOC-SRS-31, SDOC-SRS-101, scope=file)
"""

from typing import Any, Optional, Tuple

from strictdoc.backend.sdoc.models.grammar_element import ReferenceType
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.mid import MID


@auto_described
class FileEntry:
    def __init__(
        self,
        parent: Any,
        g_file_format: Optional[str],
        g_file_path: str,
        g_line_range: Optional[str],
        function: Optional[str],
        clazz: Optional[str],
    ):
        self.parent = parent

        # Default: FileEntryFormat.SOURCECODE  # noqa: ERA001
        self.g_file_format: Optional[str] = g_file_format

        # TODO: Might be worth to prohibit the use of Windows paths altogether.
        self.g_file_path: str = g_file_path
        file_path_posix = g_file_path.replace("\\", "/")
        self.file_path_posix = file_path_posix

        # The textX parser passes an empty string even if there is no LINE_RANGE
        # provided in the SDoc source.
        g_line_range = (
            g_line_range
            if g_line_range is not None and len(g_line_range) > 0
            else None
        )

        self.g_line_range: Optional[str] = g_line_range
        self.line_range: Optional[Tuple[int, int]] = None
        if g_line_range is not None:
            range_components_str = g_line_range.split(", ")
            assert len(range_components_str) == 2, range_components_str
            self.line_range = (
                int(range_components_str[0]),
                int(range_components_str[1]),
            )

        # The textX parser parses an optional element as an empty string. We
        # make it to None ourselves.
        self.function: Optional[str] = (
            function if function is not None and len(function) > 0 else None
        )
        self.clazz: Optional[str] = (
            clazz if clazz is not None and len(clazz) > 0 else None
        )


class FileEntryFormat:
    SOURCECODE = "Sourcecode"
    PYTHON = "Python"


@auto_described
class Reference:
    def __init__(self, ref_type: str, parent: Any):
        self.parent = parent
        self.ref_type: str = ref_type
        self.role: Optional[str] = None


@auto_described
class FileReference(Reference):
    def __init__(
        self, parent: Any, g_file_entry: FileEntry, role: Optional[str] = None
    ):
        super().__init__(ReferenceType.FILE, parent)
        self.g_file_entry: FileEntry = g_file_entry
        self.role: Optional[str] = (
            role if role is not None and len(role) > 0 else None
        )
        self.mid = MID.create()

    def get_posix_path(self) -> str:
        return self.g_file_entry.file_path_posix

    def get_file_format(self) -> Optional[str]:
        return self.g_file_entry.g_file_format


@auto_described
class ParentReqReference(Reference):
    def __init__(self, parent: Any, ref_uid: str, role: Optional[str]):
        super().__init__(ReferenceType.PARENT, parent)
        self.ref_uid: str = ref_uid
        # When ROLE: field is not provided for a parent reference, the
        # textX still passes relation_uid as an empty string (instead of None
        # as one could expect).
        self.role: Optional[str] = (
            role if role is not None and len(role) > 0 else None
        )
        self.mid = MID.create()


@auto_described
class ChildReqReference(Reference):
    def __init__(self, parent: Any, ref_uid: str, role: str):
        super().__init__(ReferenceType.CHILD, parent)
        self.ref_uid = ref_uid
        # When ROLE: field is not provided for a child reference, the
        # textX still passes relation_uid as an empty string (instead of None
        # as one could expect).
        self.role: Optional[str] = (
            role if role is not None and len(role) > 0 else None
        )
        self.mid = MID.create()



from typing import Any, List, Optional


class FreeTextContainer:
    def __init__(self, parts: List[Any]) -> None:
        assert isinstance(parts, list)
        self.parts = parts
        self.ng_level = None
        self.ng_line_start: Optional[int] = None
        self.ng_line_end: Optional[int] = None
        self.ng_col_start: Optional[int] = None
        self.ng_col_end: Optional[int] = None
        self.ng_byte_start: Optional[int] = None
        self.ng_byte_end: Optional[int] = None



from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.document_tree_iterator import DocumentTreeIterator
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.core.traceability_index_builder import TraceabilityIndexBuilder
from strictdoc.git.project_diff_analyzer import (
    ChangeStats,
    ProjectDiffAnalyzer,
    ProjectTreeDiffStats,
)
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.parallelizer import NullParallelizer


class ChangeContainer:
    def __init__(
        self,
        *,
        change_stats: ChangeStats,
        documents_iterator_lhs: DocumentTreeIterator,
        documents_iterator_rhs: DocumentTreeIterator,
        lhs_stats: ProjectTreeDiffStats,
        rhs_stats: ProjectTreeDiffStats,
        traceability_index_lhs: TraceabilityIndex,
        traceability_index_rhs: TraceabilityIndex,
    ):
        self.change_stats: ChangeStats = change_stats
        self.documents_iterator_lhs = documents_iterator_lhs
        self.documents_iterator_rhs = documents_iterator_rhs
        self.lhs_stats: ProjectTreeDiffStats = lhs_stats
        self.rhs_stats: ProjectTreeDiffStats = rhs_stats
        self.traceability_index_lhs: TraceabilityIndex = traceability_index_lhs
        self.traceability_index_rhs: TraceabilityIndex = traceability_index_rhs


class ChangeGenerator:
    @staticmethod
    def generate(
        lhs_project_config: ProjectConfig,
        rhs_project_config: ProjectConfig,
    ) -> ChangeContainer:
        assert isinstance(lhs_project_config, ProjectConfig)
        assert isinstance(rhs_project_config, ProjectConfig)

        parallelizer = NullParallelizer()

        traceability_index_lhs: TraceabilityIndex = (
            TraceabilityIndexBuilder.create(
                project_config=lhs_project_config,
                parallelizer=parallelizer,
                # We don't want to deal with source files for Diff.
                skip_source_files=True,
            )
        )

        traceability_index_rhs: TraceabilityIndex = (
            TraceabilityIndexBuilder.create(
                project_config=rhs_project_config,
                parallelizer=parallelizer,
                # We don't want to deal with source files for Diff.
                skip_source_files=True,
            )
        )

        lhs_stats: ProjectTreeDiffStats = (
            ProjectDiffAnalyzer.analyze_document_tree(traceability_index_lhs)
        )
        rhs_stats: ProjectTreeDiffStats = (
            ProjectDiffAnalyzer.analyze_document_tree(traceability_index_rhs)
        )
        change_stats: ChangeStats = ChangeStats.create_from_two_indexes(
            traceability_index_lhs, traceability_index_rhs, lhs_stats, rhs_stats
        )

        documents_iterator_lhs = DocumentTreeIterator(
            assert_cast(traceability_index_lhs.document_tree, DocumentTree)
        )
        documents_iterator_rhs = DocumentTreeIterator(
            assert_cast(traceability_index_rhs.document_tree, DocumentTree)
        )

        return ChangeContainer(
            change_stats=change_stats,
            documents_iterator_lhs=documents_iterator_lhs,
            documents_iterator_rhs=documents_iterator_rhs,
            lhs_stats=lhs_stats,
            rhs_stats=rhs_stats,
            traceability_index_lhs=traceability_index_lhs,
            traceability_index_rhs=traceability_index_rhs,
        )



from dataclasses import dataclass
from typing import Union

from markupsafe import Markup

from strictdoc import __version__
from strictdoc.backend.sdoc.models.anchor import Anchor
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.document_tree import DocumentTree
from strictdoc.core.document_tree_iterator import DocumentTreeIterator
from strictdoc.core.file_tree import File, Folder
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.html.generators.view_objects.helpers import (
    screen_should_display_file,
    screen_should_display_folder,
)
from strictdoc.export.html.html_templates import JinjaEnvironment
from strictdoc.export.html.renderers.link_renderer import LinkRenderer
from strictdoc.helpers.cast import assert_cast


@dataclass
class ProjectTreeViewObject:
    def __init__(
        self,
        *,
        traceability_index: TraceabilityIndex,
        project_config: ProjectConfig,
    ):
        self.traceability_index: TraceabilityIndex = traceability_index
        self.project_config: ProjectConfig = project_config

        link_renderer = LinkRenderer(
            root_path="", static_path=project_config.dir_for_sdoc_assets
        )
        self.link_renderer: LinkRenderer = link_renderer

        self.document_tree_iterator: DocumentTreeIterator = (
            DocumentTreeIterator(
                assert_cast(traceability_index.document_tree, DocumentTree)
            )
        )
        self.standalone: bool = False
        self.is_running_on_server: bool = project_config.is_running_on_server
        self.strictdoc_version = __version__
        self.contains_included_documents = (
            traceability_index.contains_included_documents
        )

    def render_local_anchor(
        self, node: Union[Anchor, SDocNode, SDocSection, SDocDocument]
    ) -> str:
        return self.link_renderer.render_local_anchor(node)

    def render_screen(self, jinja_environment: JinjaEnvironment) -> Markup:
        return jinja_environment.render_template_as_markup(
            "screens/project_index/index.jinja", view_object=self
        )

    def render_map(self, jinja_environment: JinjaEnvironment) -> Markup:
        return jinja_environment.render_template_as_markup(
            "screens/project_index/project_map.jinja.js", view_object=self
        )

    def render_static_url(self, url: str) -> str:
        return self.link_renderer.render_static_url(url)

    def render_url(self, url: str) -> str:
        return self.link_renderer.render_url(url)

    def render_static_url_with_prefix(self, url: str) -> str:
        return self.link_renderer.render_static_url_with_prefix(url)

    def is_empty_tree(self) -> bool:
        return self.document_tree_iterator.is_empty_tree()

    def get_document_level(self) -> int:
        return 0

    def should_display_fragments_toggle(self) -> bool:
        return self.project_config.export_included_documents

    def should_display_folder(self, folder: Folder) -> bool:
        return screen_should_display_folder(
            folder,
            self.traceability_index,
            self.project_config,
            must_only_include_non_included_sdoc=False,
        )

    def should_display_file(self, file: File) -> bool:
        return screen_should_display_file(
            file,
            self.traceability_index,
            self.project_config,
            must_only_include_non_included_sdoc=False,
        )



import datetime
import glob
import hashlib
import os.path
import shutil
from pathlib import Path
from typing import Any, List, Optional

from jinja2 import (
    Environment,
    FileSystemLoader,
    ModuleLoader,
    StrictUndefined,
    Template,
)
from markupsafe import Markup

from strictdoc import environment
from strictdoc.core.project_config import ProjectConfig
from strictdoc.export.html.jinja.assert_extension import AssertExtension
from strictdoc.helpers.file_modification_time import get_file_modification_time
from strictdoc.helpers.timing import measure_performance


class JinjaEnvironment:
    environment: Environment

    def __init__(self, environment: Environment):
        self.environment = environment

    def get_template(self, *args: Any, **kwargs: Any) -> Template:
        return self.environment.get_template(*args, **kwargs)

    def render_template_as_markup(
        self, template: str, *args: Any, **kwargs: Any
    ) -> Markup:
        return Markup(
            self.environment.get_template(template).render(*args, **kwargs)
        )


class HTMLTemplates:
    @staticmethod
    def create(
        project_config: ProjectConfig,
        enable_caching: bool,
        strictdoc_last_update: datetime.datetime,
    ) -> "HTMLTemplates":
        assert isinstance(strictdoc_last_update, datetime.datetime)
        if enable_caching:
            cacheable_templates = CompiledHTMLTemplates(project_config)
            cacheable_templates.reset_jinja_environment_if_outdated(
                strictdoc_last_update
            )
            cacheable_templates.compile_jinja_templates()
            return CompiledHTMLTemplates(project_config)

        return NormalHTMLTemplates()

    def jinja_environment(self) -> JinjaEnvironment:
        raise NotImplementedError


class NormalHTMLTemplates(HTMLTemplates):
    def __init__(self) -> None:
        self._jinja_environment: JinjaEnvironment = JinjaEnvironment(
            Environment(
                loader=FileSystemLoader(
                    environment.get_path_to_html_templates()
                ),
                undefined=StrictUndefined,
                extensions=[AssertExtension],
                autoescape=True,
            )
        )

    def jinja_environment(self) -> JinjaEnvironment:
        return self._jinja_environment


class CompiledHTMLTemplates(HTMLTemplates):
    def __init__(self, project_config: ProjectConfig):
        path_to_output_dir_hash = hashlib.md5(
            project_config.output_dir.encode("utf-8")
        ).hexdigest()
        self.path_to_jinja_cache_bucket_dir = os.path.join(
            project_config.get_path_to_cache_dir(),
            "jinja",
            path_to_output_dir_hash,
        )
        self._jinja_environment: Optional[JinjaEnvironment] = None

    def compile_jinja_templates(self) -> None:
        if os.path.isdir(self.path_to_jinja_cache_bucket_dir):
            return
        jinja_environment = Environment(
            loader=FileSystemLoader(environment.get_path_to_html_templates()),
            undefined=StrictUndefined,
            extensions=[AssertExtension],
            autoescape=True,
        )
        # TODO: Check if this line is still needed (might be some older workaround).
        jinja_environment.globals.update(isinstance=isinstance)
        with measure_performance("Compile Jinja templates"):

            def filter_function_(name: str) -> bool:
                # On macOS, the .DS_Store files make Jinja templates compiler
                # to crash.
                # https://github.com/strictdoc-project/strictdoc/issues/1266
                if name.endswith(".DS_Store"):
                    return False
                return True

            Path(self.path_to_jinja_cache_bucket_dir).mkdir(
                parents=True, exist_ok=True
            )
            jinja_environment.compile_templates(
                self.path_to_jinja_cache_bucket_dir,
                zip=None,
                filter_func=filter_function_,
                ignore_errors=False,
            )

    def jinja_environment(self) -> JinjaEnvironment:
        if self._jinja_environment is not None:
            return self._jinja_environment
        assert os.path.isdir(self.path_to_jinja_cache_bucket_dir)
        self._jinja_environment = JinjaEnvironment(
            Environment(
                loader=ModuleLoader(self.path_to_jinja_cache_bucket_dir),
                undefined=StrictUndefined,
                extensions=[AssertExtension],
                autoescape=True,
            )
        )
        return self._jinja_environment

    def reset_jinja_environment_if_outdated(
        self, strictdoc_last_update: datetime.datetime
    ) -> None:
        assert isinstance(strictdoc_last_update, datetime.datetime)

        if os.path.isdir(self.path_to_jinja_cache_bucket_dir):
            jinja_cache_files: List[str] = list(
                glob.iglob(
                    f"{self.path_to_jinja_cache_bucket_dir}/**/*.py",
                    recursive=True,
                )
            )

            jinja_cache_mtime = get_file_modification_time(jinja_cache_files[0])

            if strictdoc_last_update > jinja_cache_mtime:
                self._jinja_environment = None
                shutil.rmtree(self.path_to_jinja_cache_bucket_dir)



"""
@relation(SDOC-SRS-143, scope=file)
"""

import os
from enum import IntEnum
from typing import Any, List, Optional, Type, Union

import bs4
from bs4 import BeautifulSoup

from strictdoc.backend.sdoc.document_reference import DocumentReference
from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.document_grammar import DocumentGrammar
from strictdoc.backend.sdoc.models.node import SDocNode, SDocNodeField
from strictdoc.backend.sdoc.models.reference import (
    FileEntry,
    FileEntryFormat,
    FileReference,
)
from strictdoc.backend.sdoc.models.section import SDocSection
from strictdoc.core.file_tree import File
from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.cast import assert_cast, assert_optional_cast
from strictdoc.helpers.paths import path_to_posix_path


class JUnitXMLFormat(IntEnum):
    LLVM_LIT = 1
    CTEST = 2
    GOOGLE_TEST = 3
    PYTEST = 4

    @staticmethod
    def create_from_path(file_path: str) -> "JUnitXMLFormat":
        if file_path.endswith(".lit.junit.xml"):
            return JUnitXMLFormat.LLVM_LIT
        if file_path.endswith(".ctest.junit.xml"):
            return JUnitXMLFormat.CTEST
        if file_path.endswith(".gtest.junit.xml"):
            return JUnitXMLFormat.GOOGLE_TEST
        if file_path.endswith(".pytest.junit.xml"):
            return JUnitXMLFormat.PYTEST
        raise NotImplementedError(file_path)


class JUnitXMLReader:
    @classmethod
    def read_from_file(
        cls: Type["JUnitXMLReader"],
        doc_file: File,
        project_config: ProjectConfig,
    ) -> SDocDocument:
        with open(doc_file.full_path, encoding="UTF-8") as file:
            content = file.read()
        return cls.read_from_string(content, doc_file, project_config)

    @classmethod
    def read_from_string(
        cls: Type["JUnitXMLReader"],
        content: str,
        doc_file: File,
        project_config: ProjectConfig,
    ) -> SDocDocument:
        if len(content) == 0:
            raise RuntimeError(
                "Document is empty, line 1, column 1 (<string>, line 1)"
            )
        try:
            soup = BeautifulSoup(content, "xml")

        except Exception as exception:  # pylint: disable=broad-except
            raise RuntimeError(str(exception)) from None

        xml_format = JUnitXMLFormat.create_from_path(doc_file.full_path)

        document = SDocDocument(
            mid=None,
            title="Test report",
            config=None,
            view=None,
            grammar=None,
            section_contents=[],
        )
        document.ng_including_document_reference = DocumentReference()
        grammar = DocumentGrammar.create_for_test_report(document)
        document.grammar = grammar
        document.config.requirement_style = "Table"

        xml_testsuite_list: List[
            Union[
                bs4.element.PageElement,
                bs4.element.Tag,
                bs4.element.NavigableString,
            ]
        ]
        xml_testsuites: Optional[bs4.element.Tag] = assert_optional_cast(
            soup.find("testsuites", recursive=False), bs4.element.Tag
        )

        # Some tools, e.g., LLVM LIT, produce JUnit XML files with a top-level
        # <testsuites> tag which allows containing multiple <testsuite> tags.
        # Some tools, e.g., Google Test, produce only a single <testsuite>.
        if xml_testsuites is not None:
            assert isinstance(xml_testsuites, bs4.element.Tag)
            xml_testsuite_list = xml_testsuites.find_all(
                "testsuite", recursive=False
            )
        else:
            xml_testsuite: bs4.element.Tag = assert_cast(
                soup.find("testsuite", recursive=False), bs4.element.Tag
            )
            xml_testsuite_list = [xml_testsuite]
        if len(xml_testsuite_list) == 1:
            assert isinstance(xml_testsuite_list[0], bs4.element.Tag)
            document.title = "Test report: " + assert_cast(
                xml_testsuite_list[0]["name"], str
            )

        for xml_testsuite_ in xml_testsuite_list:
            assert isinstance(xml_testsuite_, bs4.element.Tag)
            title = assert_cast(xml_testsuite_["name"], str)
            total_tests: int = int(str(xml_testsuite_["tests"]))
            total_failures: int = int(str(xml_testsuite_["failures"]))
            total_skipped: int = int(str(xml_testsuite_["skipped"]))
            total_success: int = total_tests - total_failures - total_skipped

            test_suite_section = SDocSection(
                parent=document,
                mid=None,
                uid=None,
                custom_level=None,
                title=title,
                requirement_prefix=None,
                section_contents=[],
            )
            test_suite_section.ng_including_document_reference = (
                DocumentReference()
            )
            test_suite_section.ng_document_reference = DocumentReference()
            test_suite_section.ng_document_reference.set_document(document)
            document.section_contents.append(test_suite_section)

            summary_table = f"""\
.. list-table:: Test suite summary
    :widths: 25 10
    :header-rows: 0

    * - **Number of tests:**
      - {total_tests}
    * - **Number of successful tests:**
      - {total_success}
    * - **Number of failed tests:**
      - {total_failures}
    * - **Number of skipped tests:**
      - {total_skipped}
"""

            testcase_node = SDocNode(
                parent=test_suite_section,
                node_type="TEXT",
                fields=[],
                relations=[],
            )
            testcase_node.ng_document_reference = DocumentReference()
            testcase_node.ng_document_reference.set_document(document)
            testcase_node.ng_including_document_reference = DocumentReference()
            testcase_node.set_field_value(
                field_name="STATEMENT",
                form_field_index=0,
                value=SDocNodeField(
                    parent=testcase_node,
                    field_name="STATEMENT",
                    parts=[summary_table],
                    multiline__="True",
                ),
            )
            test_suite_section.section_contents.append(testcase_node)

            #
            # Parse individual <testcase> elements.
            #

            xml_testcase_list: List[Any] = xml_testsuite_.find_all(
                "testcase", recursive=False
            )
            for xml_testcase_ in xml_testcase_list:
                assert isinstance(xml_testcase_, bs4.element.Tag)

                test_case_node_uid: str
                test_case_node_title: str
                test_case_node_status: str = "PASSED"
                test_case_node_test_path: Optional[str] = None
                test_case_node_test_function: Optional[str] = None

                xml_testcase_name: str = assert_cast(xml_testcase_["name"], str)
                xml_testcase_classname: str = assert_cast(
                    xml_testcase_["classname"], str
                )
                xml_testcase_time: str = assert_cast(xml_testcase_["time"], str)
                xml_error_or_none: Optional[bs4.element.Tag] = (
                    assert_optional_cast(
                        xml_testcase_.find("error"), bs4.element.Tag
                    )
                )
                xml_failure_or_none: Optional[bs4.element.Tag] = (
                    assert_optional_cast(
                        xml_testcase_.find("failure"), bs4.element.Tag
                    )
                )
                xml_skipped_or_none: Optional[bs4.element.Tag] = (
                    assert_optional_cast(
                        xml_testcase_.find("skipped"), bs4.element.Tag
                    )
                )

                if xml_error_or_none is not None:
                    raise RuntimeError(
                        "JUnit XML contains a test that failed due to an error: "
                        f"{xml_testcase_name}"
                    )

                if xml_failure_or_none is not None:
                    test_case_node_status = "FAILED"
                elif xml_skipped_or_none is not None:
                    test_case_node_status = "SKIPPED"

                #
                # Different tools produce different outputs when it comes to how
                # the test names and paths are stored. Each tool's output is
                # handled separately below.
                #
                if xml_format == JUnitXMLFormat.LLVM_LIT:
                    #
                    # Example produced by LLVM LIT:
                    # <testcase classname="StrictDoc integration tests.tests/integration" name="test.ignored.itest" time="5.50"/>
                    #

                    rel_path_to_test_suite = (
                        project_config.test_report_root_dict.get(
                            doc_file.rel_path.relative_path_posix, None
                        )
                    )

                    if rel_path_to_test_suite is None:
                        raise RuntimeError(
                            "The relative path to the test suite must be "
                            "registered in the strictdoc.toml config under the "
                            "'test_report_root_dict' option: "
                            f"{doc_file.rel_path.relative_path_posix}"
                        )

                    # Relative path to test is a combination of the classname and name,
                    # but we must remove the test suite name prefix.
                    rel_path_to_test: str = os.path.join(
                        xml_testcase_classname, xml_testcase_name
                    )
                    _, _, rel_path_to_test = rel_path_to_test.partition(".")

                    rel_path_to_test = os.path.join(
                        rel_path_to_test_suite, rel_path_to_test
                    )
                    rel_path_to_test = path_to_posix_path(rel_path_to_test)

                    assert project_config.source_root_path is not None
                    full_path_to_test = os.path.join(
                        project_config.source_root_path, rel_path_to_test
                    )
                    assert os.path.isfile(full_path_to_test), full_path_to_test

                    test_case_node_uid = rel_path_to_test
                    test_case_node_title = rel_path_to_test
                    test_case_node_duration = xml_testcase_time
                    test_case_node_test_path = rel_path_to_test
                elif xml_format == JUnitXMLFormat.CTEST:
                    test_case_node_uid = xml_testcase_name
                    test_case_node_title = xml_testcase_name
                    test_case_node_duration = xml_testcase_time
                    test_case_node_test_function = "#GTEST#" + xml_testcase_name
                elif xml_format == JUnitXMLFormat.GOOGLE_TEST:
                    google_test_name = (
                        xml_testcase_classname + "." + xml_testcase_name
                    )
                    test_case_node_uid = google_test_name
                    test_case_node_title = google_test_name
                    test_case_node_duration = xml_testcase_time
                    test_case_node_test_function = "#GTEST#" + google_test_name
                elif xml_format == JUnitXMLFormat.PYTEST:
                    xml_testcase_path_parts = xml_testcase_classname
                    xml_testcase_class = ""
                    # Heuristic: if last part of the classname attribute starts with "Test" and is
                    # in CamelCase, then we assume it is the test class name:
                    # - Remove it from the path.
                    # - Prepend it to the function name.
                    parts = xml_testcase_classname.split(".")
                    if (
                        parts
                        and parts[-1].startswith("Test")
                        and parts[-1][1:].lower() != parts[-1][1:]
                    ):
                        xml_testcase_path_parts = ".".join(parts[:-1])
                        xml_testcase_class = parts[-1] + "."
                    test_case_node_uid = (
                        xml_testcase_classname + "." + xml_testcase_name
                    )
                    test_case_node_duration = xml_testcase_time
                    test_case_node_test_path = (
                        xml_testcase_path_parts.replace(".", os.path.sep)
                        + ".py"
                    )
                    test_case_node_title = xml_testcase_classname
                    test_case_node_test_function = (
                        xml_testcase_class + xml_testcase_name
                    )
                else:
                    raise NotImplementedError("Unsupported JUnit XML format")

                testcase_node = SDocNode(
                    parent=test_suite_section,
                    node_type="TEST_RESULT",
                    fields=[],
                    relations=[],
                )
                testcase_node.ng_document_reference = DocumentReference()
                testcase_node.ng_document_reference.set_document(document)
                testcase_node.ng_including_document_reference = (
                    DocumentReference()
                )
                if xml_skipped_or_none is None:
                    testcase_node.set_field_value(
                        field_name="UID",
                        form_field_index=0,
                        value=SDocNodeField(
                            parent=testcase_node,
                            field_name="UID",
                            parts=[test_case_node_uid],
                            multiline__=None,
                        ),
                    )
                if test_case_node_test_path is not None:
                    testcase_node.set_field_value(
                        field_name="TEST_PATH",
                        form_field_index=0,
                        value=SDocNodeField(
                            parent=testcase_node,
                            field_name="TEST_PATH",
                            parts=[
                                path_to_posix_path(test_case_node_test_path)
                            ],
                            multiline__=None,
                        ),
                    )
                if test_case_node_test_function is not None:
                    testcase_node.set_field_value(
                        field_name="TEST_FUNCTION",
                        form_field_index=0,
                        value=SDocNodeField(
                            parent=testcase_node,
                            field_name="TEST_FUNCTION",
                            parts=[test_case_node_test_function],
                            multiline__=None,
                        ),
                    )
                testcase_node.set_field_value(
                    field_name="DURATION",
                    form_field_index=0,
                    value=SDocNodeField(
                        parent=testcase_node,
                        field_name="DURATION",
                        parts=[test_case_node_duration],
                        multiline__=None,
                    ),
                )
                testcase_node.set_field_value(
                    field_name="STATUS",
                    form_field_index=0,
                    value=SDocNodeField(
                        parent=testcase_node,
                        field_name="STATUS",
                        parts=[test_case_node_status],
                        multiline__=None,
                    ),
                )
                testcase_node.set_field_value(
                    field_name="TITLE",
                    form_field_index=0,
                    value=SDocNodeField(
                        parent=testcase_node,
                        field_name="TITLE",
                        parts=[test_case_node_title],
                        multiline__=None,
                    ),
                )
                if xml_skipped_or_none is None:
                    if test_case_node_test_path is not None:
                        testcase_node.relations.append(
                            FileReference(
                                parent=testcase_node,
                                g_file_entry=FileEntry(
                                    parent=None,
                                    g_file_format=FileEntryFormat.SOURCECODE,
                                    g_file_path=test_case_node_test_path,
                                    g_line_range=None,
                                    function=test_case_node_test_function,
                                    clazz=None,
                                ),
                            )
                        )
                    elif test_case_node_test_function is not None:
                        testcase_node.relations.append(
                            FileReference(
                                parent=testcase_node,
                                g_file_entry=FileEntry(
                                    parent=None,
                                    g_file_format=FileEntryFormat.SOURCECODE,
                                    g_file_path="#FORWARD#",
                                    g_line_range=None,
                                    function=test_case_node_test_function,
                                    clazz=None,
                                ),
                            )
                        )
                test_suite_section.section_contents.append(testcase_node)

        return document



REGEX_ROLE = r"[^)]+"

SOURCE_FILE_GRAMMAR = f"""
SourceFileTraceabilityInfo[noskipws]:
  g_parts += Part
;

Part[noskipws]:
  // The EmptyLine is needed in addition to the SingleLineString because
  // otherwise textX's get_location() ignores the whitespaces.
  // TODO: Maybe there is a trick to disable that and only use SingleLineString.
  EmptyLine | RangeMarker | LineMarker | FunctionRangeMarker | SingleLineString
;

EmptyLine[noskipws]:
  '\n'
;

RangeMarker[noskipws]:
  // It is a hard-won result: it is important that the "@sdoc" is within the
  // regex. Putting it next to the regex as "@sdoc" does not work.
  // TODO: It would be great to check this with the TextX developers.
  (
  /^.*?@sdoc/
  (begin_or_end = "[/" | begin_or_end = "[")
  (reqs_objs += Req[', ']) ']' '\n'?
  )
  |
  (
  /^.*?@relation/
  '('
  (reqs_objs += Req[', ']) ', scope=' scope=/(range_start|range_end)/ (", role=" role=/{REGEX_ROLE}/)? ')' '\n'?
  )
;

FunctionRangeMarker[noskipws]:
  /^.*?@relation/
  "(" (reqs_objs += Req[', ']) ', scope=' scope="file" (", role=" role=/{REGEX_ROLE}/)? ')' '\n'?
;

LineMarker[noskipws]:
  // It is a hard-won result: it is important that the "@sdoc" is within the
  // regex. Putting it next to the regex as "@sdoc" does not work.
  // TODO: It would be great to check this with the TextX developers.
  (
  /^.*?@sdoc/
  "(" (reqs_objs += Req[', ']) ')' '\n'?
  )
  |
  (
  /^.*?@relation/
  "(" (reqs_objs += Req[', ']) ', scope=line' (", role=" role=/{REGEX_ROLE}/)? ")" '\n'?
  )
;

Req[noskipws]:
  uid = /(?!scope=)[A-Za-z][A-Za-z0-9\\-]+/
;

SingleLineString[noskipws]:
  !RangeMarker /.+/ '\n'?
;
"""



# This function could be a separate package but keeping it within the project
# for simplicity.
__version__ = "0.0.1"

from typing import Any, Callable, Optional, Type, TypeVar, Union, overload

T = TypeVar("T", bound=Type[Any])


@overload
def auto_described(cls: T) -> T: ...
@overload
def auto_described(
    cls: None = ..., *, str_and_repr: bool = ...
) -> Callable[[T], T]: ...


# Maybe there is a better way to generate __str__ and __repr__.
# But for now, this solution works good enough:
# https://stackoverflow.com/a/33800620/598057
# https://stackoverflow.com/a/24617244/598057
def auto_described(
    cls: Optional[Any] = None, str_and_repr: bool = True
) -> Union[T, Callable[[T], T]]:
    def configure_class(clz: T) -> T:
        def __str__(self: object) -> str:
            return auto_str(self)

        def __repr__(self: object) -> str:
            return auto_str(self)

        clz.__str__ = __str__  # type: ignore[assignment]
        if str_and_repr:
            clz.__repr__ = __repr__  # type: ignore[assignment]
        return clz

    if cls is not None:
        return configure_class(cls)

    def factory(clz: T) -> T:
        return configure_class(clz)

    return factory


def auto_str(obj: object) -> str:
    items = []

    # This is a rather defensive implementation that prevents auto_str from
    # breaking on the objects with recursive references:
    # Example: A -> B -> A.
    for prop, value in obj.__dict__.items():
        if isinstance(value, list):
            if len(value) == 0:
                item = f"{prop} = []"
            else:
                item = f"{prop} = [{len(value)} elements]"
        elif isinstance(value, dict):
            if len(value) == 0:
                item = f"{prop} = {{}}"
            else:
                item = f"{prop} = {{{len(value)} elements}}"
        elif isinstance(value, set):
            item = f"{prop} = {value.__class__.__name__}({len(value)} elements)"
        elif isinstance(value, str):
            item = f'{prop} = "{value}"'
        elif isinstance(value, bytes):
            item = f"{prop} = {value!r}"
        elif isinstance(value, (int, float, bool)):
            item = f"{prop} = {value}"
        elif isinstance(value, object):
            item = f"{prop} = {value.__class__.__name__}(...)"
        else:
            item = f"{prop} = {value}"

        items.append(item)
    return f"{obj.__class__.__name__}({', '.join(items)})"



from typing import Any, Dict, List, Optional, Union

from typing_extensions import TypeAlias

from strictdoc.backend.sdoc_source_code.models.function import Function
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.line_marker import LineMarker
from strictdoc.backend.sdoc_source_code.models.range_marker import (
    ForwardRangeMarker,
    RangeMarker,
)
from strictdoc.backend.sdoc_source_code.models.source_node import SourceNode
from strictdoc.core.source_tree import SourceFile
from strictdoc.helpers.auto_described import auto_described
from strictdoc.helpers.file_stats import SourceFileStats

RelationMarkerType: TypeAlias = Union[
    FunctionRangeMarker, LineMarker, RangeMarker, ForwardRangeMarker
]


@auto_described
class SourceFileTraceabilityInfo:
    """
    Class that keeps all traceability info related to a single source file.

    At the init time, only the backward RangeMarkers are available from
    a source file. At runtime, the ForwardRangeMarkers are mixed in
    from the Requirement/FileReference links. This is why the .markers
    is a union.

    Note that g_parts is only used by the textX grammar to allow the
    definition of the child grammar elements. All other instance variables
    of this class are not using or based on g_parts. Instead, they are
    populated during the textX processing step, e.g., via
    general_language_marker_processors.py.
    """

    def __init__(self, g_parts: List[Any]):  # noqa: ARG002
        self.source_file: Optional[SourceFile] = None
        self.source_nodes: List[SourceNode] = []
        self.functions: List[Function] = []

        #
        # {                                              # noqa: ERA001
        #  "REQ-001": [RangeMarker(...), ...],           # noqa: ERA001
        #  "REQ-002": [RangeMarker(...), ...],           # noqa: ERA001
        # }                                              # noqa: ERA001
        self.ng_map_reqs_to_markers: Dict[str, List[RelationMarkerType]] = {}

        self.ng_map_names_to_markers: Dict[str, List[RelationMarkerType]] = {}
        self.ng_map_names_to_definition_functions: Dict[str, Function] = {}

        #
        # Merged ranges contain ranges that are fully covered by one or more
        # forward or backward relations. If a range is part of a larger range,
        # it gets merged into the larger range. The merged ranges are tracked so
        # that for each source code function it can be answered whether the
        # function is covered by any requirement or not.
        #
        self.merged_ranges: List[List[int]] = []
        self.file_stats: SourceFileStats = SourceFileStats()
        self.ng_lines_covered = 0
        self.lines_info: Dict[int, bool] = {}
        self._coverage: float = 0
        self.covered_functions: int = 0

        self.markers: List[RelationMarkerType] = []

    # FIXME: is_requirement() will go away.
    def is_requirement(self) -> bool:
        return False

    # FIXME: is_section() will go away.
    def is_section(self) -> bool:
        return False

    # FIXME: is_section() will go away.
    def is_source_file(self) -> bool:
        return True

    def get_coverage(self) -> float:
        return self._coverage

    def set_coverage_stats(
        self,
        merged_ranges: List[List[int]],
        lines_covered: int,
    ) -> None:
        self.merged_ranges = merged_ranges
        self.ng_lines_covered = lines_covered
        self._coverage = (
            round(lines_covered / self.file_stats.lines_non_empty * 100, 1)
            if self.file_stats.lines_non_empty != 0
            else 0
        )



"""
@relation(SDOC-SRS-18, scope=file)
"""

from abc import ABC, abstractmethod
from typing import Generator, List, Optional, Union

from strictdoc.backend.sdoc.models.document_config import (
    DocumentConfig,
)
from strictdoc.backend.sdoc_source_code.models.source_file_info import (
    SourceFileTraceabilityInfo,
)
from strictdoc.core.document_meta import DocumentMeta
from strictdoc.helpers.mid import MID


class RequirementFieldName:
    MID = "MID"
    UID = "UID"
    PREFIX = "PREFIX"
    LEVEL = "LEVEL"
    STATUS = "STATUS"
    TAGS = "TAGS"
    TITLE = "TITLE"

    # {STATEMENT, DESCRIPTION, CONTENT} are aliases.
    # It is assumed that either field is provided for each node.
    STATEMENT = "STATEMENT"
    DESCRIPTION = "DESCRIPTION"
    CONTENT = "CONTENT"

    RATIONALE = "RATIONALE"
    COMMENT = "COMMENT"


RESERVED_NON_META_FIELDS = [
    RequirementFieldName.TITLE,
    RequirementFieldName.STATEMENT,
    RequirementFieldName.DESCRIPTION,
    RequirementFieldName.CONTENT,
    RequirementFieldName.COMMENT,
    RequirementFieldName.RATIONALE,
    RequirementFieldName.LEVEL,
]


class SDocNodeFieldIF(ABC):
    parent: "SDocNodeIF"


class SDocNodeIF(ABC):
    reserved_mid: MID
    mid_permanent: bool
    parent: Union["SDocNodeIF", "SDocDocumentIF", "SDocSectionIF"]
    node_type: str
    section_contents: List["SDocElementIF"]
    ng_level: Optional[int]
    ng_resolved_custom_level: Optional[str]
    ng_whitelisted: bool
    ng_has_requirements: bool

    @property
    @abstractmethod
    def reserved_uid(self) -> Optional[str]:
        raise NotImplementedError

    @reserved_uid.setter
    @abstractmethod
    def reserved_uid(self, uid: Optional[str]) -> None:
        raise NotImplementedError

    @property
    def reserved_title(self) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def is_normative_node(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def is_text_node(self) -> bool:
        raise NotImplementedError

    @abstractmethod
    def get_debug_info(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_document(self) -> Optional["SDocDocumentIF"]:
        raise NotImplementedError

    @abstractmethod
    def get_including_document(self) -> Optional["SDocDocumentIF"]:
        raise NotImplementedError

    @abstractmethod
    def get_parent_or_including_document(self) -> "SDocDocumentIF":
        raise NotImplementedError

    @abstractmethod
    def get_prefix(self) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def blacklist_if_needed(self) -> None:
        raise NotImplementedError


class SDocSectionIF(ABC):
    reserved_mid: MID
    reserved_uid: Optional[str]
    mid_permanent: bool
    parent: Union["SDocDocumentIF", "SDocSectionIF"]
    section_contents: List["SDocElementIF"]
    ng_level: Optional[int]
    ng_resolved_custom_level: Optional[str]
    ng_whitelisted: bool
    ng_has_requirements: bool

    @abstractmethod
    def get_document(self) -> Optional["SDocDocumentIF"]:
        raise NotImplementedError

    @abstractmethod
    def get_prefix(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_debug_info(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_display_title(self, include_toc_number: bool = True) -> str:
        raise NotImplementedError

    @abstractmethod
    def blacklist_if_needed(self) -> None:
        raise NotImplementedError


class SDocGrammarIF:
    pass


class SDocDocumentIF(ABC):
    reserved_mid: MID
    mid_permanent: bool
    section_contents: List["SDocElementIF"]
    included_documents: List["SDocDocumentIF"]
    config: DocumentConfig
    grammar: Optional[SDocGrammarIF]
    meta: Optional[DocumentMeta]
    is_bundle_document: bool
    ng_level: Optional[int]
    ng_whitelisted: bool
    ng_has_requirements: bool

    # FIXME: Get rid of @property everywhere.
    @property
    def reserved_uid(self) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def get_prefix(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def get_debug_info(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def iterate_included_documents_depth_first(
        self,
    ) -> Generator["SDocDocumentIF", None, None]:
        raise NotImplementedError

    @abstractmethod
    def get_display_title(self, include_toc_number: bool = True) -> str:
        raise NotImplementedError

    @property
    def ng_resolved_custom_level(self) -> Optional[str]:
        raise NotImplementedError

    @abstractmethod
    def blacklist_if_needed(self) -> None:
        raise NotImplementedError


class SDocDocumentFromFileIF(ABC):
    parent: Union[SDocDocumentIF, SDocSectionIF]
    ng_resolved_custom_level: Optional[str]
    ng_whitelisted: bool

    @abstractmethod
    def iterate_nodes(
        self,
        element_type: Optional[str] = None,
    ) -> Generator[SDocNodeIF, None, None]:
        raise NotImplementedError

    @property
    def section_contents(self) -> List[SDocDocumentIF]:
        raise NotImplementedError


SDocElementIF = Union[
    SDocNodeIF,
    SDocSectionIF,
    SDocDocumentIF,
    SDocDocumentFromFileIF,
]


SDocIteratedElementIF = Union[
    SDocNodeIF,
    SDocSectionIF,
    SDocDocumentIF,
]


SDocExtendedElementIF = Union[
    SDocElementIF,
    SourceFileTraceabilityInfo,
]



# pragma: no cover file.

"""
Experimental code to generate SPDX.

This code is excluded from code coverage calculation because the feature is
highly experimental and most of this code will be removed.
"""

import os.path
import re
from datetime import datetime
from pathlib import Path
from typing import Dict, Optional, Union

from semantic_version import Version
from spdx_tools.spdx3.model import (
    Element,
    Hash,
    HashAlgorithm,
    ProfileIdentifierType,
    Relationship,
    RelationshipType,
    SpdxDocument,
)
from spdx_tools.spdx3.model.positive_integer_range import PositiveIntegerRange
from spdx_tools.spdx3.model.software import (
    File,
    Package,
    Snippet,
    SoftwarePurpose,
)
from spdx_tools.spdx3.model.spdx_document import CreationInfo
from spdx_tools.spdx3.payload import Payload
from spdx_tools.spdx3.writer.console.relationship_writer import (
    write_relationship,
)
from spdx_tools.spdx3.writer.console.software.file_writer import write_file
from spdx_tools.spdx3.writer.console.software.package_writer import (
    write_package,
)
from spdx_tools.spdx3.writer.console.software.snippet_writer import (
    write_snippet,
)
from spdx_tools.spdx3.writer.console.spdx_document_writer import (
    write_spdx_document,
)
from spdx_tools.spdx3.writer.json_ld.json_ld_writer import write_payload

from strictdoc.backend.sdoc.models.document import SDocDocument
from strictdoc.backend.sdoc.models.node import SDocNode
from strictdoc.backend.sdoc.models.reference import (
    ParentReqReference,
)
from strictdoc.backend.sdoc.writer import SDWriter
from strictdoc.core.document_iterator import DocumentCachingIterator
from strictdoc.core.project_config import ProjectConfig
from strictdoc.core.traceability_index import TraceabilityIndex
from strictdoc.export.spdx.spdx_sdoc_container import SPDXSDocContainer
from strictdoc.export.spdx.spdx_to_sdoc_converter import SPDXToSDocConverter
from strictdoc.helpers.cast import assert_cast
from strictdoc.helpers.sha256 import get_sha256

RELATION_ID_HOW_TO = "SPDXRef-Relationship-How-to-form-ID?"


def create_relationship_summary(
    lhs: Element, rhs: Element, relation: str
) -> str:
    return f"{lhs.summary} --|{relation}|--> {rhs.summary}"


def get_spdx_ref(node: Union[SDocDocument, SDocNode, str]) -> str:
    if isinstance(node, str):
        return re.sub(r"[/\\ ]", "_", node)

    identifier: Optional[str] = None
    if node.mid_permanent:
        identifier = node.reserved_mid
    elif node.reserved_uid is not None:
        identifier = node.reserved_uid
    elif node.reserved_title is not None:
        identifier = re.sub(r"[ #-/\\]", "_", node.reserved_title)
    else:
        raise RuntimeError(
            f"Cannot create unique identifier for this node: {node}"
        )

    if isinstance(node, SDocDocument):
        return "SDocDocument-" + identifier
    if isinstance(node, SDocNode):
        return "SDocRequirement-" + identifier
    raise NotImplementedError


class SDocToSPDXConverter:
    @staticmethod
    def create_document(project_config: ProjectConfig) -> SpdxDocument:
        spec_version = Version("3.0.0")
        creation_info = CreationInfo(
            spec_version=spec_version,
            created=datetime.today(),
            created_by=[],
            profile=[ProfileIdentifierType.SOFTWARE],
            data_license="CC0 1.0",
            created_using=["SPDXRef-StrictDoc"],
            comment=f"SPDX 3.0 SBOM for {project_config.project_title}'s requirements.",
        )
        spdx_document = SpdxDocument(
            spdx_id="SPDXRef-DOCUMENT",
            name=f"Requirements for {project_config.project_title}",
            element=[],
            root_element=[],
            creation_info=creation_info,
            summary=f"SPDX Document for project {project_config.project_title}",
            description=None,
            comment=None,
        )
        return spdx_document

    @staticmethod
    def create_package(project_config: ProjectConfig) -> Package:
        return Package(
            spdx_id="SPDXRef-PACKAGE",
            name="Requirements package",
            summary=f"SPDX Package for project {project_config.project_title}",
            description=None,
            comment=None,
            verified_using=[
                Hash(
                    algorithm=HashAlgorithm.SHA256,
                    hash_value="TBD: What to calculate for a package?",
                )
            ],
            primary_purpose=SoftwarePurpose.BOM,
            homepage=None,
        )

    @staticmethod
    def create_document_to_file(
        document: SDocDocument, document_bytes: bytes
    ) -> File:
        assert document.meta is not None

        return File(
            spdx_id=f"SPDXRef-File-{get_spdx_ref(document)}",
            name=document.meta.document_filename,
            summary=f"SDPX File for document {document.title}",
            description=None,
            comment=None,
            verified_using=[
                Hash(
                    algorithm=HashAlgorithm.SHA256,
                    hash_value=get_sha256(document_bytes),
                )
            ],
            primary_purpose=SoftwarePurpose.DOCUMENTATION,
        )

    @staticmethod
    def convert_file_to_file(node_link_path: str, file_bytes: bytes) -> File:
        return File(
            spdx_id=f"SPDXRef-File-{get_spdx_ref(node_link_path)}",
            name=node_link_path,
            summary=f"SPDX File for source file {node_link_path}",
            description=None,
            comment=None,
            verified_using=[
                Hash(
                    algorithm=HashAlgorithm.SHA256,
                    hash_value=get_sha256(file_bytes),
                )
            ],
            primary_purpose=SoftwarePurpose.SOURCE,
        )

    @staticmethod
    def convert_requirement_to_snippet(
        requirement: SDocNode, document_bytes: bytes, spdx_file: File
    ) -> Snippet:
        snippet_sha256 = get_sha256(
            document_bytes[requirement.ng_byte_start : requirement.ng_byte_end]
        )
        assert requirement.reserved_uid is not None
        assert requirement.reserved_statement is not None

        return Snippet(
            spdx_id=requirement.reserved_uid,
            primary_purpose=SoftwarePurpose.DOCUMENTATION,
            name=f"Requirement '{requirement.reserved_title}'",
            summary=f"SPDX Snippet for requirement {requirement.reserved_uid}",
            description=requirement.reserved_statement.partition("\n")[0],
            comment=(
                "This snippet has been generated from a requirement "
                f"defined in a StrictDoc file: {spdx_file.name}."
            ),
            verified_using=[
                Hash(
                    algorithm=HashAlgorithm.SHA256,
                    hash_value=snippet_sha256,
                )
            ],
            byte_range=PositiveIntegerRange(
                requirement.ng_byte_start, requirement.ng_byte_end
            ),
            line_range=PositiveIntegerRange(
                requirement.ng_line_start, requirement.ng_line_end
            ),
        )


class SPDXGenerator:
    @staticmethod
    def export_tree(
        project_config: ProjectConfig,
        traceability_index: TraceabilityIndex,
        output_spdx_root: str,
    ) -> None:
        Path(output_spdx_root).mkdir(parents=True, exist_ok=True)

        sdoc_spdx_converter: SDocToSPDXConverter = SDocToSPDXConverter()
        spdx_container: SPDXSDocContainer = SPDXSDocContainer()

        #
        # SPDX Document and SPDX Package
        #
        spdx_document = sdoc_spdx_converter.create_document(project_config)
        spdx_container.document = spdx_document

        spdx_package = sdoc_spdx_converter.create_package(project_config)
        spdx_container.package = spdx_package

        spdx_document.root_element = [spdx_package.spdx_id]

        spdx_container.relationships.append(
            Relationship(
                spdx_id=RELATION_ID_HOW_TO,
                from_element=spdx_document.spdx_id,
                relationship_type=RelationshipType.CONTAINS,
                to=[spdx_package.spdx_id],
                name=None,
                summary=create_relationship_summary(
                    spdx_document,
                    spdx_package,
                    "CONTAINS",
                ),
                description=None,
                comment=None,
            )
        )

        lookup_uid_to_requirement_snippet: Dict[str, Snippet] = {}
        lookup_file_name_to_spdx_file: Dict[str, File] = {}

        for document_ in traceability_index.document_tree.document_list:
            assert document_.meta is not None
            with open(document_.meta.input_doc_full_path, "rb") as input_file_:
                document_bytes = input_file_.read()

            #
            # Create SPDX File from SDoc Document.
            #
            spdx_file = sdoc_spdx_converter.create_document_to_file(
                document_, document_bytes
            )
            spdx_container.files.append(spdx_file)
            spdx_container.relationships.append(
                Relationship(
                    spdx_id=RELATION_ID_HOW_TO,
                    from_element=spdx_package.spdx_id,
                    relationship_type=RelationshipType.CONTAINS,
                    to=[spdx_file.spdx_id],
                    summary=create_relationship_summary(
                        spdx_package,
                        spdx_file,
                        "CONTAINS",
                    ),
                )
            )
            spdx_container.map_spdx_ref_to_objects[spdx_file.spdx_id] = (
                spdx_file
            )

            document_iterator: DocumentCachingIterator = (
                traceability_index.get_document_iterator(document_)
            )

            for node, _ in document_iterator.all_content():
                if (
                    isinstance(node, SDocNode)
                    and node.node_type == "REQUIREMENT"
                ):
                    if node.reserved_uid is None:
                        continue

                    assert node.reserved_title is not None, (
                        "The current implementation only supports requirements with a title."
                    )

                    #
                    # Create SPDX Snippet from SDoc Requirement.
                    #
                    spdx_snippet: Snippet = (
                        sdoc_spdx_converter.convert_requirement_to_snippet(
                            node, document_bytes, spdx_file
                        )
                    )
                    lookup_uid_to_requirement_snippet[node.reserved_uid] = (
                        spdx_snippet
                    )
                    spdx_container.snippets.append(spdx_snippet)
                    spdx_container.map_spdx_ref_to_objects[
                        spdx_snippet.spdx_id
                    ] = spdx_snippet
                    spdx_container.map_spdx_snippets_to_files[
                        spdx_snippet.spdx_id
                    ] = spdx_file.spdx_id

                    spdx_container.relationships.append(
                        Relationship(
                            spdx_id=RELATION_ID_HOW_TO,
                            from_element=spdx_file.spdx_id,
                            relationship_type=RelationshipType.CONTAINS,
                            to=[spdx_snippet.spdx_id],
                            name=None,
                            summary=create_relationship_summary(
                                spdx_file,
                                spdx_snippet,
                                "CONTAINS",
                            ),
                            description=None,
                            comment=None,
                        )
                    )

                    node_links = traceability_index.get_requirement_file_links(
                        node
                    )
                    for node_link_path_, _range_markers in node_links:
                        if node_link_path_ in lookup_file_name_to_spdx_file:
                            continue

                        with open(node_link_path_, "rb") as node_link_file_:
                            file_bytes = node_link_file_.read()

                        source_spdx_file = (
                            sdoc_spdx_converter.convert_file_to_file(
                                node_link_path_, file_bytes
                            )
                        )
                        lookup_file_name_to_spdx_file[node_link_path_] = (
                            source_spdx_file
                        )
                        spdx_container.map_spdx_ref_to_objects[
                            source_spdx_file.spdx_id
                        ] = source_spdx_file

                        spdx_container.files.append(source_spdx_file)
                        spdx_container.relationships.append(
                            Relationship(
                                spdx_id=RELATION_ID_HOW_TO,
                                from_element=spdx_snippet.spdx_id,
                                relationship_type=RelationshipType.REQUIREMENT_FOR,
                                to=[source_spdx_file.spdx_id],
                                name=None,
                                summary=create_relationship_summary(
                                    spdx_snippet,
                                    source_spdx_file,
                                    "REQUIREMENT_FOR",
                                ),
                                description=None,
                                comment=None,
                            )
                        )

        #
        # Now iterate over requirement nodes again.
        #
        for document_ in traceability_index.document_tree.document_list:
            document_iterator = traceability_index.get_document_iterator(
                document_
            )
            for node, _ in document_iterator.all_content():
                if not isinstance(node, SDocNode):
                    continue

                if node.is_requirement():
                    if node.reserved_uid is None:
                        continue

                requirement: SDocNode = assert_cast(node, SDocNode)
                assert requirement.reserved_uid is not None
                requirement_snippet: Snippet = (
                    lookup_uid_to_requirement_snippet[requirement.reserved_uid]
                )

                for reference_ in requirement.relations:
                    if isinstance(reference_, ParentReqReference):
                        parent_requirement_snippet: Snippet = (
                            lookup_uid_to_requirement_snippet[
                                reference_.ref_uid
                            ]
                        )

                        spdx_container.relationships.append(
                            Relationship(
                                spdx_id=RELATION_ID_HOW_TO,
                                from_element=requirement_snippet.spdx_id,
                                relationship_type=RelationshipType.REQUIREMENT_FOR,
                                to=[parent_requirement_snippet.spdx_id],
                                name=None,
                                summary=create_relationship_summary(
                                    requirement_snippet,
                                    parent_requirement_snippet,
                                    "REQUIREMENT_FOR",
                                ),
                                description=None,
                                comment=None,
                            )
                        )

        #
        # Output SPDX to a file.
        #

        output_path = os.path.join(output_spdx_root, "output.spdx")
        with open(output_path, "w", encoding="utf8") as file:
            write_spdx_document(spdx_document, text_output=file)
            file.write("\n")

            write_package(spdx_package, text_output=file, heading=True)
            file.write("\n")

            for file_ in spdx_container.files:
                write_file(file_, text_output=file)
                file.write("\n")

            for snippet_ in spdx_container.snippets:
                write_snippet(snippet_, text_output=file)
                file.write("\n")

            for relationship_ in spdx_container.relationships:
                write_relationship(relationship_, text_output=file)
                file.write("\n")

        if True:
            payload: Payload = Payload()

            payload.add_element(spdx_document)
            payload.add_element(spdx_package)
            for spdx_file_ in spdx_container.files:
                payload.add_element(spdx_file_)
            for spdx_snippet_ in spdx_container.snippets:
                payload.add_element(spdx_snippet_)

            path_to_output_spdx_json = os.path.join(
                output_spdx_root, "output.spdx"
            )

            write_payload(payload, path_to_output_spdx_json)

        if True:
            sdoc_document = SPDXToSDocConverter.convert(spdx_container)

            sdoc_output: str = SDWriter(project_config).write(sdoc_document)

            sdoc_output_path = os.path.join(
                output_spdx_root, "output.spdx.sdoc"
            )
            with open(sdoc_output_path, "w", encoding="utf8") as output_file_:
                output_file_.write(sdoc_output)



"""
@relation(SDOC-SRS-142, scope=file)
"""

from typing import Any, List, Set

from strictdoc.backend.sdoc_source_code.constants import FunctionAttribute
from strictdoc.backend.sdoc_source_code.models.function_range_marker import (
    FunctionRangeMarker,
)
from strictdoc.helpers.auto_described import auto_described


@auto_described
class Function:
    def __init__(
        self,
        parent: Any,
        name: str,
        display_name: str,
        line_begin: int,
        line_end: int,
        child_functions: List[Any],
        markers: List[FunctionRangeMarker],
        attributes: Set[FunctionAttribute],
    ):
        assert parent is not None
        self.parent = parent
        self.name = name
        self.display_name = display_name

        # Child functions are supported in programming languages that can nest
        # functions, for example, Python.
        self.child_functions: List[Function] = child_functions
        self.markers: List[FunctionRangeMarker] = markers
        self.line_begin = line_begin
        self.line_end = line_end
        self.attributes: Set[FunctionAttribute] = attributes

    def is_declaration(self) -> bool:
        return FunctionAttribute.DECLARATION in self.attributes

    def is_definition(self) -> bool:
        return FunctionAttribute.DEFINITION in self.attributes

    def is_public(self) -> bool:
        return FunctionAttribute.STATIC not in self.attributes



"""
A simple container class for tracking dependencies between artifacts.

It provides answers to whether a given artifact has to be re-generated or not.

@relation(SDOC-SRS-2, scope=file)
"""

import datetime
import os.path
from copy import deepcopy
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Optional, Set

from strictdoc.core.project_config import ProjectConfig
from strictdoc.helpers.file_modification_time import (
    get_file_modification_time,
)
from strictdoc.helpers.pickle import pickle_dump, pickle_load


@dataclass
class FileDependencyManager:
    path_to_cache_dir: str
    path_to_cache: str
    dependencies_now: Dict[str, Set[str]]
    dependencies_prev: Dict[str, Set[str]]
    dependencies_must_renegerate: Set[str]

    @staticmethod
    def create_from_cache(
        project_config: ProjectConfig,
    ) -> "FileDependencyManager":
        path_to_cache_dir = project_config.get_path_to_cache_dir()
        path_to_cache = os.path.join(path_to_cache_dir, "dependencies")

        if not os.path.isfile(path_to_cache):
            return FileDependencyManager(
                path_to_cache_dir, path_to_cache, {}, {}, set()
            )

        with open(path_to_cache, "rb") as cache_file:
            cache_file_bytes = cache_file.read()

        unpickled_content: Optional[FileDependencyManager] = pickle_load(
            cache_file_bytes
        )

        if unpickled_content is not None:
            assert isinstance(unpickled_content, FileDependencyManager), (
                unpickled_content,
            )
            unpickled_content.dependencies_now.clear()
            return unpickled_content

        raise AssertionError(  # pragma: no cover
            f"Problem reading the file dependency cache file: {path_to_cache}"
        )

    def must_generate(self, path_to_input_file: str) -> bool:
        must_regenerate = (
            path_to_input_file in self.dependencies_must_renegerate
        )
        return must_regenerate

    def add_dependency(
        self,
        path_to_file: str,
        path_to_dependent_file: str,
    ) -> None:
        dependencies = self.dependencies_now.setdefault(path_to_file, set())
        dependencies.add(path_to_dependent_file)

    def save_to_cache(self) -> None:
        self.dependencies_prev = deepcopy(self.dependencies_now)

        pickled_content = pickle_dump(self)

        Path(self.path_to_cache_dir).mkdir(exist_ok=True, parents=True)
        with open(self.path_to_cache, "wb") as cache_file:
            cache_file.write(pickled_content)

    def resolve_modification_dates(
        self, strictdoc_last_update: datetime.datetime
    ) -> None:
        self.dependencies_must_renegerate.clear()

        items_now, items_before = (
            self.dependencies_now.items(),
            self.dependencies_prev.items(),
        )
        for items_ in [items_now, items_before]:
            for path_to_input_file_, dependencies_ in items_:
                for path_to_dependency_ in dependencies_:
                    if (
                        # The file has not been generated yet.
                        not os.path.isfile(path_to_dependency_)
                        # The file used to exist but not anymore.
                        or not os.path.isfile(path_to_input_file_)
                        # The file is outdated compared to its HTML output artifact.
                        or get_file_modification_time(path_to_input_file_)
                        > get_file_modification_time(path_to_dependency_)
                        # The file is outdated compared to StrictDoc's own code (this
                        # branch is development-only).
                        or strictdoc_last_update
                        > get_file_modification_time(path_to_dependency_)
                    ):
                        self.dependencies_must_renegerate.add(
                            path_to_dependency_
                        )

        self.save_to_cache()
